/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.aed;

import java.util.Map;

/**
 * A listener for the {@link Topic} that will be informed of events.
 */
public interface TopicListener
{
	/**
	 * Called on the successful creation of a topic. Triggered as a success
	 * response as a result of the following method(s) being previously called
	 * by this user: {@link AED#createTopic(String, TopicListener)},
	 * {@link AED#createTopic(String, int, TopicListener)}.
	 *
	 * @param topic topic object returning the message.
	 *
	 * @param data Contains the key value pairs that this topic consists of.
	 */
	void onTopicConnected(Topic topic, Map<String, Object> data);

	/**
	 * Called as a result of not being able to connect to a topic after calling
	 * connect on this object. Triggered as a failure response as a result of
	 * the following method(s) being previously called by this user:
	 * {@link AED#createTopic(String, TopicListener)},
	 * {@link AED#createTopic(String, int, TopicListener)}.
	 *
	 * @param topic topic object returning the message
	 *
	 * @param message message returning the state of the operation
	 *
	 */
	void onTopicNotConnected(Topic topic, String message);

	/**
	 * Called as a result of topic successfully deleted. Triggered as a success
	 * response as a result of the following method(s) being previously called
	 * by this user: {@link Topic#disconnect(boolean delete)} passing
	 * <code>true</code> for the delete argument.
	 *
	 * @param topic topic object returning the message.
	 *
	 * @param message message string that contains details on the successful
	 *            delete.
	 */
	void onTopicDeleted(Topic topic, String message);

	/**
	 * Called as a result of a topic not being deleted. Triggered as a failure
	 * response as a result of the following method being previously called by
	 * this user: {@link Topic#disconnect(boolean delete)} passing
	 * <code>true</code> for the delete argument.
	 *
	 * @param topic topic object returning the message
	 *
	 * @param message message returning the state of the operation
	 */
	void onTopicNotDeleted(Topic topic, String message);

	/**
	 * Called when Data successfully submitted. Triggered as a success response
	 * as a result of the following method being previously called by this
	 * user: {@link Topic#submitData(String key, String value)}.
	 *
	 * @param topic topic object returning the message.
	 *
	 * @param key a unique identifier which is associated with a value.
	 *
	 * @param value the single value that is associated to the specified key.
	 *
	 * @param version the version of the change - it should be incremented by
	 *            one for any change that is made to this key-value pair.
	 */
	void onTopicSubmitted(Topic topic, String key, String value, int version);

	/**
	 * Called as a result of data not being submitted. Triggered as a failure
	 * response as a result of the following method being previously called by
	 * this user: {@link Topic#submitData(String key, String value)}.
	 *
	 * @param topic topic object returning the message
	 *
	 * @param key a unique identifier which is associated with a value.
	 *
	 * @param value the single value that is associated to the specified key.
	 *
	 * @param message message returning the state of the operation, giving more
	 *            detail on the error.
	 */
	void onTopicNotSubmitted(Topic topic, String key, String value,
			String message);

	/**
	 * Called on Data successfully deleted. Triggered as a success response as
	 * a result of the following method being previously called by this user:
	 * {@link Topic#deleteData(String)}.
	 *
	 * @param topic topic object returning the message.
	 *
	 * @param key a unique identifier which is associated with a value.
	 *
	 * @param version the version of the change - it should be incremented by
	 *            one for any change that is made to this key-value pair.
	 */
	void onDataDeleted(Topic topic, String key, int version);

	/**
	 * Called as a result of a topic's data failing to delete after a request
	 * for it to do so. Triggered as a failure response as a result of the
	 * following method(s) being previously called by this user:
	 * {@link Topic#deleteData(String)}.
	 *
	 * @param topic topic object returning the message
	 *
	 * @param key a unique identifier which is associated with a value.
	 *
	 * @param message message returning the state of the operation, giving more
	 *            detail on the error.
	 */
	void onDataNotDeleted(Topic topic, String key, String message);

	/**
	 * Called as a result of message successfully sent. Triggered as a success
	 * response as a result of the following method being previously called by
	 * this user: {@link Topic#sendAedMessage(String)}.
	 *
	 * @param topic topic object returning the message.
	 *
	 * @param message message returning the state of the operation.
	 */
	void onTopicSent(Topic topic, String message);

	/**
	 * Called as a result of a message send failure. Triggered as a failure
	 * response as a result of the following method(s) being previously called
	 * by this user: {@link Topic#sendAedMessage(String)}.
	 *
	 * @param topic topic object returning the message
	 *
	 * @param message the message that the connected user was trying to send
	 * via the sendAedMessage.
	 *
	 * @param error message returning the state of the operation, giving more
	 *            detail on the error.
	 */
	void onTopicNotSent(Topic topic, String message, String error);

	/**
	 * Called as a result of the topic expiring, or being deleted by another
	 * user that is connected to the topic. Triggered as a success response as
	 * a result of the following method being previously called by any user:
	 * {@link Topic#disconnect(boolean delete)} passing <code>true</code> for
	 * the delete argument.
	 *
	 * @param topic topic object returning the message
	 */
	void onTopicDeletedRemotely(Topic topic);

	/**
	 * Called when the topic is updated, by one of the users connected to the
	 * topic calling submitData on the Topic object. Triggered as a success
	 * response as a result of the following method being previously called by
	 * any user: {@link Topic#submitData(String key, String value)}.
	 *
	 * @param topic topic object returning the message
	 *
	 * @param key a unique identifier which is associated with a value
	 *
	 * @param value a value that is retrieved via the use of the unique key.
	 *
	 * @param version the version of the change - it should be incremented by
	 *            one for any change that is made to this key-value pair.
	 *
	 * @param deleted whether or not the data has been deleted.
	 *
	 */
    void onTopicUpdated(Topic topic, String key, String value, int version, boolean deleted);

	/**
	 * Called as a result of message successfully received. Triggered as a
	 * success response as a result of the following method(s) being previously
	 * called by any user: {@link Topic#sendAedMessage(String)}.
	 *
	 * @param topic topic object returning the message
	 *
	 * @param message message that was received.
	 */
	void onMessageReceived(Topic topic, String message);
}
