/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.impl;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.TrustManager;

import org.json.JSONException;
import org.json.JSONObject;

public class ServerMessageManager implements ServerMessageTransportListener
{
    private final Map<String, ServerMessageHandler> handlers = new HashMap<String, ServerMessageHandler>();
    private TransportFailureHandler transportFailureHandler;
    private ServerMessageTransport transport;
    private String transportUrl;
    private JSONObject startMessage;

    public ServerMessageManager(ServerMessageTransport transport, String session)
    {
        super();
        this.transport = transport;
        this.transportUrl = decodeSessionId(session);
    }

    public void addHandler(String messageType, ServerMessageHandler handler)
    {
        handlers.put(messageType, handler);
    }

    public void setNetworkReachable(boolean reachable)
    {
        this.transport.setNetworkReachable(reachable);
    }

    /**
     * Opens the transport and sends an the given message when the transport
     * informs us that it's open. If the transport is already open, the message
     * is sent immediately.
     * 
     * @param transportFailureHandler The transport failure handler.
     * @param json The message to send once the transport is open.
     */
    public void startWithServerMessage(TransportFailureHandler transportFailureHandler, JSONObject json)
    {
        this.transportFailureHandler = transportFailureHandler;
        this.startMessage = json;
        this.transport.open(transportUrl);
    }

    public void sendServerMessage(JSONObject json)
    {
        this.transport.sendServerMessage(json);
    }

    public void stop()
    {
        this.transport.close();
        // TODO: Which handler should we use for this? We need to close all
        // calls, conversations and topics in this case.
    }

    @Override
    public void onTransportOpened()
    {
        if (startMessage != null)
        {
            sendServerMessage(startMessage);
        }
    }

    @Override
    public void onServerMessage(JSONObject json)
    {
        try
        {
            final String type = json.getString(ServerMessageBuilder.MESSAGE_FIELD_TYPE);
            final ServerMessageHandler handler = handlers.get(type);
            if (handler != null)
            {
                handler.handleServerMessage(json);
            }
            else
            {
                // TODO: LOG THIS
            }
        }
        catch (JSONException ex)
        {
            // TODO: log this
            ex.printStackTrace();
        }
    }

    @Override
    public void onTransportFailed()
    {
        if (transportFailureHandler != null)
        {
            transportFailureHandler.handleTransportFailure();
        }
    }

    /**
     * Decode the given session ID with a simple character swapping algorithm.
     * 
     * @param encodedKey
     * @return
     */
    private static String decodeSessionId(final String encodedKey)
    {
        final String base = "/:?#[]@%$&'()*+,;=._~- ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890";
        final String target = "!:?#[]@%$&'()*+,;=._~- 0987654321ZYXWVUTSRQPONMLKJIHGFEDCBAzyxwvutsrqponmlkjihgfedcba";

        String key = null;
        try
        {
            key = URLDecoder.decode(encodedKey, Charset.defaultCharset().displayName());
        }
        catch (UnsupportedEncodingException e)
        {
            e.printStackTrace();
            return null;
        }

        StringBuilder cb = new StringBuilder();
        final int strLen = key.length();
        for (int i = 0; i < strLen; i++)
        {
            cb.append(base.charAt(target.indexOf(key.charAt(i))));
        }
        return cb.toString();
    }

    @Override
    public void onTransportRetry(int attemptNumber, long delayUntilNextAttempt) 
    {
        if (transportFailureHandler != null)
        {
            transportFailureHandler.handleTransportRetry(attemptNumber, delayUntilNextAttempt);
        }
        
    }

    @Override
    public void onTransportReestablished() 
    {
        if (transportFailureHandler != null)
        {
            transportFailureHandler.handleTransportReestablished();
        }
        
    }

    public void setTrustManager(TrustManager trustManager) 	 
    { 	 
        this.transport.setTrustManager(trustManager); 	 
    }
    
    public void setHostnameVerifier(HostnameVerifier hostnameVerifier)
    {
    	this.transport.setHostnameVerifier(hostnameVerifier);
    }
    
    public void setUseCookies(final boolean useCookies)
    {
        this.transport.setUseCookies(useCookies);
    }
}
