/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone;

import java.util.List;

import android.content.Context;
import android.graphics.Point;
import android.view.Surface;

import com.alicecallsbob.fcsdk.android.ListenerTarget;

/**
 * An object that acts as the entry point for all voice and video calls.
 */
public interface Phone extends ListenerTarget<PhoneListener>
{
    /**
     * Returns all calls that are currently in progress.
     *
     * @return all calls that are currently in progress.
     */
    List<? extends Call> getCurrentCalls();


    /**
     * Create a VideoSurface instance for rendering local or remote video.
     *
     * @param c The application Context
     * @param screenDimensions Point
     * @param listener A VideoSurfaceListener
     * @return videosurface
     */
    VideoSurface createVideoSurface(Context c, Point screenDimensions, VideoSurfaceListener listener);

    /**
     * Set the VideoSurface that renders the local video stream.
     *
     * @param view the VideoSurface to render the local video stream.
     */
    void setPreviewView(VideoSurface view);

    /**
     * Sets the orientation of the video being sent. Constants for the parameter
     * are defined in android.view.Surface.ROTATION_xxx.
     *
     * @param orientation The orientation of the video to be sent
     * @see Surface#ROTATION_0
     * @see Surface#ROTATION_90
     * @see Surface#ROTATION_180
     * @see Surface#ROTATION_270
     */
    void setVideoOrientation(int orientation);

    /**
     * Sets the camera index to use as the video source. Changing the camera takes effect immediately, even during an
     * active call.
     * <p>
     * If the SDK user does not call this operation, the SDK defaults to the first front facing camera it can find. 
     * If there is no front facing camera, the index of the first rear facing camera is returned.
     *
     * @param camera The camera index to use as the video source.
     */
    void setCamera(int camera);
    
    /**
     * Enables or disables local streams.
     *
     * @param isAudioEnabled Whether local audio should be enabled.
     */
    void enableLocalAudio(boolean isAudioEnabled);

    /**
     * Enables or disables local streams.
     *
     * @param isVideoEnabled Whether local video should be enabled.
     */
    void enableLocalVideo(boolean isVideoEnabled);

    /**
     * Creates a call to the given remote address.
     * <p>
     * Note that the SDK only supports one call at a time. If you attempt to create 
     * a call whilst another is already in pregress, then the call creation will fail.
     *
     * @param address The remote address.
     * @param audio Whether to use audio.
     * @param video Whether to use video.
     * @param listener The call listener.
     * @return The new call
     * @throws CallCreationWithErrorException if call creation failed
     */
    Call createCall(String address, boolean audio, boolean video, CallListener listener)
        throws CallCreationWithErrorException;
    
    /**
     * Returns the recommended video capture settings for the selected camera.
     * 
     * @return The recommended video capture settings
     */
    List<PhoneVideoCaptureSetting> getRecommendedCaptureSettings();
    
    /**
     * Returns the recommended video capture settings for the specified camera.
     * 
     * @return The recommended video capture settings
     */
    List<PhoneVideoCaptureSetting> getRecommendedCaptureSettings(int camera);   
    
    /**
     * Sets the preferred video captured resolution.
     * 
     * NOTE: a different resolution may be selected if the preferred resolution cannot be used.
     * 
     * @param resolution The preferred video capture resolution
     */
    void setPreferredCaptureResolution(PhoneVideoCaptureResolution resolution);
    
    /**
     * Sets the preferred video capture framerate.
     * 
     * NOTE: a different framework may be selected if the preferred framerate is not supported
     * @param frameRate
     */
    void setPreferredCaptureFrameRate(int frameRate);
    
    /**
     * Returns the preferred capture resolution.
     * 
     * @return The preferred capture resolution
     */
    PhoneVideoCaptureResolution getPreferredCaptureResolution();
    
    /**
     * Returns the preferred capture framerate.
     * 
     * @return the preferred framerate
     */
    int getPreferredCaptureFrameRate();
    
    /**
     * Returns the user selected camera index being used for capturing the video source.
     * <p>
     * If the SDK user has not called {@link #setCamera(int)}, the SDK defaults to the first front
     * facing camera it can find. If there is no front facing camera, the index of the first rear facing camera is
     * returned.
     * 
     * @return the selected camera index as an int.
     */
    int getCamera();
    
    /**
     * Returns the AudioDeviceManager.
     * 
     * This is used to query and set the available devices for a call.
     * @return teh audio device manager in use
     */
    AudioDeviceManager getAudioDeviceManager();
}
