/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import com.alicecallsbob.fcsdk.android.phone.AudioDeviceManager;
import com.alicecallsbob.fcsdk.android.phone.AudioDeviceManagerListener;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.media.AudioManager;
import android.util.Log;

public final class AudioDeviceManagerImpl implements AudioDeviceManager
{

    /* Receiver which handles changes in wired headset availability. */
    private class WiredHeadsetReceiver extends BroadcastReceiver
    {
        /** State for wired headset unplugged */
        private static final int STATE_UNPLUGGED = 0;
        /** State for wired headset plugged in */
        private static final int STATE_PLUGGED = 1;
        /** State for wired headset has no microphone */
        private static final int HAS_NO_MIC = 0;
        /** State for wired headset has microphone */
        private static final int HAS_MIC = 1;

        @Override
        public void onReceive(Context context, Intent intent)
        {
            final int state = intent.getIntExtra("state", STATE_UNPLUGGED);
            final int microphone = intent.getIntExtra("microphone", HAS_NO_MIC);
            final String name = intent.getStringExtra("name");
            Log.d(TAG, "WiredHeadsetReceiver.onReceive" + ": "
                    + "a=" + intent.getAction()
                    + ", s=" + (state == STATE_UNPLUGGED ? "unplugged" : "plugged")
                    + ", m=" + (microphone == HAS_MIC ? "mic" : "no mic")
                    + ", n=" + name + ", sb=" + isInitialStickyBroadcast());
            hasWiredHeadset = (state == STATE_PLUGGED);
            updateAudioDeviceState();
        }
    };

    /** TAG */
    private static final String TAG = AudioDeviceManagerImpl.class.getSimpleName();

    /** BluetoothDeviceManager which handles the connections and disconnections */
    private final BluetoothDeviceManager btDeviceManager;
    /** Application context */
    private final Context appContext;
    /** The Android AudioManager service */
    private AudioManager audioManager;
    /** Holder for AudioDeviceManager state */
    private AudioManagerState amState;
    /** Holder for state of wired headset */
    private boolean hasWiredHeadset = false;
    /** AudioManager mode saved when we start */
    private int savedAudioMode = AudioManager.MODE_INVALID;
    /** Speaker state saved when we start */
    private boolean savedIsSpeakerPhoneOn = false;
    /** Microphone state saved when we start */
    private boolean savedIsMicrophoneMute = false;

    /**
     * Default audio device; speaker phone for video calls or earpiece for audio only calls.
     */
    private AudioDevice defaultAudioDevice;

    /**
     * Contains the currently selected audio device. This device is changed automatically using a certain scheme where
     * e.g. a wired headset "wins" over speaker phone. It is also possible for a user to explicitly select a device (and
     * overrid any predefined scheme). See |userSelectedAudioDevice| for details.
     */
    private AudioDevice selectedAudioDevice;

    /**
     * Contains the user-selected audio device which overrides the predefined selection scheme.
     */
    private AudioDevice userSelectedAudioDevice;

    /** Broadcast receiver for wired headset intent broadcasts. */
    private BroadcastReceiver wiredHeadsetReceiver;

    /**
     * Contains a list of available audio devices. A Set collection is used to avoid duplicate elements.
     */
    private Set<AudioDevice> audioDevices = new HashSet<AudioDevice>();

    /**
     * Callback method for changes in audio focus.
     */
    private AudioManager.OnAudioFocusChangeListener audioFocusChangeListener;

    /**
     * Stores the listeners
     */
    private ArrayList<AudioDeviceManagerListener> listeners;

    private AudioDeviceManagerImpl(final Context context)
    {
        this.appContext = context;
        this.btDeviceManager = BluetoothDeviceManagerImpl.create(context, this);
        this.audioManager = (AudioManager) context.getSystemService(Context.AUDIO_SERVICE);
        this.wiredHeadsetReceiver = new WiredHeadsetReceiver();
        this.amState = AudioManagerState.UNINITIALIZED;
        this.defaultAudioDevice = AudioDevice.SPEAKER_PHONE;
        
        this.listeners = new ArrayList<AudioDeviceManagerListener>();
    }

    /**
     * Static constructor
     * @param context Application context
     * @return AudioDeviceManager
     */
    public static AudioDeviceManager create(final Context context)
    {
        return new AudioDeviceManagerImpl(context);
    }

    @Override
    public void updateAudioDeviceState()
    {
        Log.d(TAG, "--- updateAudioDeviceState: " 
                + "wired headset=" + hasWiredHeadset + ", " 
                + "BT state=" + btDeviceManager.getState());
        
        Log.d(TAG, "Device status: "
                + "available=" + audioDevices + ", "
                + "selected=" + selectedAudioDevice + ", "
                + "user selected=" + userSelectedAudioDevice);

        // Check if any Bluetooth headset is connected. The internal BT state will
        // change accordingly.
        if (btDeviceManager.deviceIsAvailable())
        {
            btDeviceManager.refreshListOfDevices();
        }

        //Find out which devices are available to use
        final Set<AudioDevice> newAudioDevices = collectNewAudioDevices();
        
        // Store state which is set to true if the device list has changed.
        boolean audioDeviceSetUpdated = !audioDevices.equals(newAudioDevices);
        // Update the existing audio device set.
        audioDevices = newAudioDevices;
        
        //Tidy up any anomalies
        correctUserSelectedDevice();
        
        // Need to start Bluetooth if it is available and user either selected it explicitly or
        // user did not select any output device.
        final boolean needBluetoothAudioStart = 
                btDeviceManager.getState() == BluetoothDeviceManager.State.DEVICE_AVAILABLE
                && (userSelectedAudioDevice == AudioDevice.NONE
                    || userSelectedAudioDevice == AudioDevice.BLUETOOTH);

        // Need to stop Bluetooth audio if user selected different device and
        // Bluetooth SCO connection is established or in the process.
        final boolean needBluetoothAudioStop =
                (btDeviceManager.getState() == BluetoothDeviceManager.State.CONNECTED 
                    || btDeviceManager.getState() == BluetoothDeviceManager.State.CONNECTING)
                && (userSelectedAudioDevice != AudioDevice.NONE
                    && userSelectedAudioDevice != AudioDevice.BLUETOOTH);

        if (btDeviceManager.getState() == BluetoothDeviceManager.State.DEVICE_AVAILABLE 
                || btDeviceManager.getState() == BluetoothDeviceManager.State.CONNECTING
            || btDeviceManager.getState() == BluetoothDeviceManager.State.CONNECTED)
        {
            Log.d(TAG, "Need BT audio: start=" + needBluetoothAudioStart + ", "
                + "stop=" + needBluetoothAudioStop + ", "
                + "BT state=" + btDeviceManager.getState());
        }

        // Start or stop Bluetooth SCO connection given states set earlier.
        if (needBluetoothAudioStop)
        {
            btDeviceManager.stopAudio();
            btDeviceManager.refreshListOfDevices();
        }

        if (needBluetoothAudioStart && !needBluetoothAudioStop)
        {
            // Attempt to start Bluetooth SCO audio (takes a few second to start).
            if (!btDeviceManager.startAudio())
            {
                // Remove BLUETOOTH from list of available devices since SCO failed.
                audioDevices.remove(AudioDevice.BLUETOOTH);
                audioDeviceSetUpdated = true;
            }
        }

        // Update selected audio device.
        AudioDevice newAudioDevice = userSelectedAudioDevice;

        if (btDeviceManager.getState() == BluetoothDeviceManager.State.CONNECTED)
        {
            // If a Bluetooth is connected, then it should be used as output audio
            // device automatically. Note that it is not sufficient that a headset is available;
            // an active SCO channel must also be up and running.
            newAudioDevice = AudioDevice.BLUETOOTH;
        }
        else if (hasWiredHeadset)
        {
            // If a wired headset is connected, but Bluetooth is not, then wired headset is used as
            // audio device.
            newAudioDevice = AudioDevice.WIRED_HEADSET;
        }
        else if (userSelectedAudioDevice == AudioDevice.NONE)
        {
            // No wired headset and no Bluetooth, hence the audio-device list can contain speaker
            // phone (on a tablet), or speaker phone and earpiece (on mobile phone).
            // |defaultAudioDevice| contains either AudioDevice.SPEAKER_PHONE or AudioDevice.EARPIECE
            // depending on the user's selection.
            newAudioDevice = defaultAudioDevice;
        }
        // Switch to new device but only if there has been any changes.
        if (newAudioDevice != selectedAudioDevice || audioDeviceSetUpdated)
        {
            // Do the required device switch.
            setAudioDeviceInternal(newAudioDevice);
            Log.d(TAG, "New device status: " + "available=" + audioDevices + ", " + "selected=" + newAudioDevice);
            if (listeners != null)
            {
                //Notify listeners
                for (AudioDeviceManagerListener listener:listeners)
                {
                    listener.onDeviceListChanged(getAudioDevices(), selectedAudioDevice);
                }
            }
        }
        Log.d(TAG, "--- updateAudioDeviceState done");
    }
    
    private Set<AudioDevice> collectNewAudioDevices()
    {
        //Update the set of available audio devices.
        final Set<AudioDevice> newAudioDevices = new HashSet<AudioDevice>();
        if (btDeviceManager.deviceIsConnected())
        {
            newAudioDevices.add(AudioDevice.BLUETOOTH);
        }

        if (hasWiredHeadset)
        {
            // If a wired headset is connected, then it is the only possible option.
            newAudioDevices.add(AudioDevice.WIRED_HEADSET);
        }
        else
        {
            // No wired headset, hence the audio-device list can contain speaker
            // phone (on a tablet), or speaker phone and earpiece (on mobile phone).
            newAudioDevices.add(AudioDevice.SPEAKER_PHONE);
            if (hasEarpiece())
            {
                newAudioDevices.add(AudioDevice.EARPIECE);
            }
        }
        return newAudioDevices;
    }
    
    private void correctUserSelectedDevice(){
        
        // Correct user selected audio devices if needed.
        if (btDeviceManager.getState() == BluetoothDeviceManager.State.DEVICE_UNAVAILABLE
                && userSelectedAudioDevice == AudioDevice.BLUETOOTH)
        {
            // If BT is not available, it can't be the user selection.
            userSelectedAudioDevice = AudioDevice.NONE;
        }
        if (hasWiredHeadset && userSelectedAudioDevice == AudioDevice.SPEAKER_PHONE)
        {
            // If user selected speaker phone, but then plugged wired headset then make
            // wired headset as user selected device.
            userSelectedAudioDevice = AudioDevice.WIRED_HEADSET;
        }
        if (!hasWiredHeadset && userSelectedAudioDevice == AudioDevice.WIRED_HEADSET)
        {
            // If user selected wired headset, but then unplugged wired headset then make
            // speaker phone as user selected device.
            userSelectedAudioDevice = AudioDevice.SPEAKER_PHONE;
        }

        Log.d(TAG, "UserSelectedAudioDevice is: " + userSelectedAudioDevice);
        
    }

    @Override
    public boolean start()
    {
        if (amState == AudioManagerState.RUNNING)
        {
            Log.e(TAG, "AudioManager is already active");
            return false;
        }
        
        amState = AudioManagerState.RUNNING;

        // Store current audio state so we can restore it when stop() is called.
        savedAudioMode = audioManager.getMode();
        savedIsSpeakerPhoneOn = audioManager.isSpeakerphoneOn();
        savedIsMicrophoneMute = audioManager.isMicrophoneMute();

        // Create an AudioManager.OnAudioFocusChangeListener instance.
        audioFocusChangeListener = new AudioManager.OnAudioFocusChangeListener()
        {
            // Called on the listener to notify if the audio focus for this listener has been changed.
            // The |focusChange| value indicates whether the focus was gained, whether the focus was lost,
            // and whether that loss is transient, or whether the new focus holder will hold it for an
            // unknown amount of time.
            @Override
            public void onAudioFocusChange(int focusChange)
            {
                String typeOfChange = "AUDIOFOCUS_NOT_DEFINED";
                switch (focusChange)
                {
                    case AudioManager.AUDIOFOCUS_GAIN:
                        typeOfChange = "AUDIOFOCUS_GAIN";
                        break;
                    case AudioManager.AUDIOFOCUS_GAIN_TRANSIENT:
                        typeOfChange = "AUDIOFOCUS_GAIN_TRANSIENT";
                        break;
                    case AudioManager.AUDIOFOCUS_GAIN_TRANSIENT_EXCLUSIVE:
                        typeOfChange = "AUDIOFOCUS_GAIN_TRANSIENT_EXCLUSIVE";
                        break;
                    case AudioManager.AUDIOFOCUS_GAIN_TRANSIENT_MAY_DUCK:
                        typeOfChange = "AUDIOFOCUS_GAIN_TRANSIENT_MAY_DUCK";
                        break;
                    case AudioManager.AUDIOFOCUS_LOSS:
                        typeOfChange = "AUDIOFOCUS_LOSS";
                        break;
                    case AudioManager.AUDIOFOCUS_LOSS_TRANSIENT:
                        typeOfChange = "AUDIOFOCUS_LOSS_TRANSIENT";
                        break;
                    case AudioManager.AUDIOFOCUS_LOSS_TRANSIENT_CAN_DUCK:
                        typeOfChange = "AUDIOFOCUS_LOSS_TRANSIENT_CAN_DUCK";
                        break;
                    default:
                        typeOfChange = "AUDIOFOCUS_INVALID";
                        break;
                }
                Log.d(TAG, "onAudioFocusChange: " + typeOfChange);
            }
        };

        // Request audio playout focus (without ducking) and install listener for changes in focus.
        final int result = audioManager.requestAudioFocus(audioFocusChangeListener, 
            AudioManager.STREAM_VOICE_CALL, AudioManager.AUDIOFOCUS_GAIN_TRANSIENT);
        if (result == AudioManager.AUDIOFOCUS_REQUEST_GRANTED)
        {
            Log.d(TAG, "Audio focus request granted for VOICE_CALL streams");
        }
        else
        {
            Log.e(TAG, "Audio focus request failed");
        }

        // Start by setting MODE_IN_COMMUNICATION as default audio mode. It is
        // required to be in this mode when playout and/or recording starts for
        // best possible VoIP performance.
        audioManager.setMode(AudioManager.MODE_IN_COMMUNICATION);

        // Always disable microphone mute during a WebRTC call.
        setMicrophoneMute(false);

        // Set initial device states.
        userSelectedAudioDevice = AudioDevice.NONE; 
        selectedAudioDevice = AudioDevice.NONE;
        audioDevices.clear();

        // Initialize and start Bluetooth if a BT device is available or initiate
        // detection of new (enabled) BT devices.
        btDeviceManager.start();

        // Do initial selection of audio device. This setting can later be changed
        // by adding/removing a BT or wired headset.
        updateAudioDeviceState();

        // Register receiver for broadcast intents related to adding/removing a
        // wired headset.
        registerReceiver(wiredHeadsetReceiver, new IntentFilter(Intent.ACTION_HEADSET_PLUG));
        Log.d(TAG, "AudioManager started");

        return true;
    }

    @Override
    public boolean stop()
    {
        if (amState != AudioManagerState.RUNNING)
        {
            Log.e(TAG, "Trying to stop AudioManager in incorrect state: " + amState);
            return false;
        }
        amState = AudioManagerState.UNINITIALIZED;

        unregisterReceiver(wiredHeadsetReceiver);

        btDeviceManager.stop();

        // Restore previously stored audio states.
        setSpeakerphoneOn(savedIsSpeakerPhoneOn);
        setMicrophoneMute(savedIsMicrophoneMute);
        audioManager.setMode(savedAudioMode);

        // Abandon audio focus. Gives the previous focus owner, if any, focus.
        audioManager.abandonAudioFocus(audioFocusChangeListener);
        audioFocusChangeListener = null;
        Log.d(TAG, "Abandoned audio focus for VOICE_CALL streams");

        listeners = null;
        Log.d(TAG, "AudioManager stopped");
        return true;
    }

    /** Gets the current earpiece state. */
    private boolean hasEarpiece()
    {
        return appContext.getPackageManager().hasSystemFeature(PackageManager.FEATURE_TELEPHONY);
    }
    
    /** Sets the microphone mute state. */
    private void setMicrophoneMute(boolean on)
    {
        final boolean wasMuted = audioManager.isMicrophoneMute();
        if (wasMuted == on)
        {
            return;
        }
        audioManager.setMicrophoneMute(on);
    }

    /** Helper method for receiver registration. */
    private void registerReceiver(BroadcastReceiver receiver, IntentFilter filter)
    {
        appContext.registerReceiver(receiver, filter);
    }
    
    /** Helper method for unregistration of an existing receiver. */
    private void unregisterReceiver(BroadcastReceiver receiver)
    {
        appContext.unregisterReceiver(receiver);
    }
    
    /** Changes selection of the currently active audio device. */
    private void setAudioDeviceInternal(AudioDevice device)
    {
      Log.d(TAG, "setAudioDeviceInternal(device=" + device + ")");
      if (!audioDevices.contains(device))
      {
          throw new AssertionError("Expected device to be in our set of devcies");
      }

      switch (device)
      {
        case SPEAKER_PHONE:
          setSpeakerphoneOn(true);
          break;
        case EARPIECE:
          setSpeakerphoneOn(false);
          break;
        case WIRED_HEADSET:
          setSpeakerphoneOn(false);
          break;
        case BLUETOOTH:
          setSpeakerphoneOn(false);
          break;
        default:
          Log.e(TAG, "Invalid audio device selection");
          break;
      }
      selectedAudioDevice = device;
    }
    
    /** Sets the speaker phone mode. */
    private void setSpeakerphoneOn(boolean on)
    {
        final boolean wasOn = audioManager.isSpeakerphoneOn();
        if (wasOn == on)
        {
            return;
        }
        audioManager.setSpeakerphoneOn(on);
    }
    
    @Override
    public void setAudioDevice(final AudioDevice userSelectedDevice)
    {
        switch (userSelectedDevice)
        {
            case SPEAKER_PHONE:
                userSelectedAudioDevice = userSelectedDevice;
                break;
            case EARPIECE:
                if (hasEarpiece())
                {
                    userSelectedAudioDevice = userSelectedDevice;
                }
                else
                {
                    userSelectedAudioDevice = AudioDevice.SPEAKER_PHONE;
                }
                break;
            case BLUETOOTH:
                userSelectedAudioDevice = userSelectedDevice;
                break;
            case WIRED_HEADSET:
                userSelectedAudioDevice = userSelectedDevice;
                break;
            case NONE: //Fallback to default
                userSelectedAudioDevice = userSelectedDevice;
                break;
            default:
                Log.e(TAG, "Invalid default audio device selection");
                break;
        }
        Log.d(TAG, "setUserSelectedAudioDevice(device=" + userSelectedAudioDevice + ")");
        updateAudioDeviceState();
    }

    @Override
    public void setDefaultAudioDevice(final AudioDevice defaultDevice)
    {
        switch (defaultDevice)
        {
            case SPEAKER_PHONE:
                defaultAudioDevice = defaultDevice;
                break;
            case EARPIECE:
                if (hasEarpiece())
                {
                    defaultAudioDevice = defaultDevice;
                }
                else
                {
                    defaultAudioDevice = AudioDevice.SPEAKER_PHONE;
                }
                break;
            default:
                Log.e(TAG, "Invalid default audio device selection");
                break;
        }
        Log.d(TAG, "setDefaultAudioDevice(device=" + defaultAudioDevice + ")");
        updateAudioDeviceState();
    }
    
    @Override
    public Set<AudioDevice> getAudioDevices()
    {
        return Collections.unmodifiableSet(new HashSet<AudioDevice>(audioDevices));
    }

    @Override
    public AudioDevice getSelectedAudioDevice()
    {
        return selectedAudioDevice;
    }

    @Override
    public void addListener(AudioDeviceManagerListener listener)
    {
        if ((listeners != null) && (listeners.contains(listener) == false))
        {
            listeners.add(listener);
        }

    }

    @Override
    public void removeListener(AudioDeviceManagerListener listener)
    {
        if (listeners != null)
        {
            listeners.remove(listener);
        }
    }
}
