/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.util.List;
import java.util.Set;

import com.alicecallsbob.fcsdk.android.phone.AudioDeviceManager;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothHeadset;
import android.bluetooth.BluetoothProfile;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.media.AudioManager;
import android.os.Handler;
import android.os.Process;
import android.util.Log;

public final class BluetoothDeviceManagerImpl implements BluetoothDeviceManager
{

    /** TAG */
    private static final String TAG = BluetoothDeviceManagerImpl.class.getSimpleName();
    
    /** Timeout in milliseconds to wait for BT device */
    private static final int BLUETOOTH_SCO_TIMEOUT_MS = 5000;
    /** Application context */
    private final Context context;
    /** BT service listener to listen for BT events */
    private final BluetoothServiceListener btServiceListener;
    /** Broadcast receiver for Intents */
    private final BroadcastReceiver broadcastReceiver;
    /** Handler for the timer */
    private final Handler handler;
    /** AudioDeviceManager to notify update to devices */
    private final AudioDeviceManager fcsdkAudioManager;
    /** Holder for the Bluetooth Adapter */
    private BluetoothAdapter btAdapter;
    /** Holder for the Bluetooth Headset proxy */
    private BluetoothHeadset btHeadset;
    /** Holder for the Bluetooth Device for lower level discovery*/
    private BluetoothDevice btDevice;
    /** Holder for the Bluetooth State */
    private State btState;
    /** Android AudioManager service*/
    private AudioManager audioManager;
    /** Counter for number of connection attempts */
    private int connectionAttempts;
    
    
    /**
     * Implementation of an interface that notifies BluetoothProfile IPC clients when they have been
     * connected to or disconnected from the service.
     */
    private class BluetoothServiceListener implements BluetoothProfile.ServiceListener
    {
        
        /** 
         * Called to notify the client when the proxy object has been connected to the service.
         * Once we have the profile proxy object, we can use it to monitor the state of the
         * connection and perform other operations that are relevant to the headset profile.*/
        @Override
        public void onServiceConnected(int profile, BluetoothProfile proxy)
        {
            
            if (profile != BluetoothProfile.HEADSET || btState == State.UNAVAILABLE)
            {
                Log.d(TAG, "Other profile " + profile);
                return;
            }
            Log.d(TAG, "BluetoothServiceListener.onServiceConnected: BT state=" + btState);
            // Android only supports one connected Bluetooth Headset at a time.
            btHeadset = (BluetoothHeadset) proxy;
            
            //Use this device or not?
            updateAudioDeviceState();
            Log.d(TAG, "onServiceConnected done: BT state=" + btState);
        }
    
        @Override
        /** Notifies the client when the proxy object has been disconnected from the service. */
        public void onServiceDisconnected(int profile)
        {
          //Might be another device disconnecting so check if we're interesting in this event
          if (profile != BluetoothProfile.HEADSET || btState == State.UNAVAILABLE)
          {
            return;
          }
          Log.d(TAG, "BluetoothServiceListener.onServiceDisconnected: BT state=" + btState);
          
          stopAudio();
          btHeadset = null;
          btDevice = null;
          btState = State.DEVICE_UNAVAILABLE;
          
          //Look for another device?
          updateAudioDeviceState();

          
          Log.d(TAG, "onServiceDisconnected done: BT state=" + btState);
        }
      }
    
    // Intent broadcast receiver which handles changes in Bluetooth device availability.
    // Detects headset changes and Bluetooth SCO state changes.
    private class BluetoothHeadsetBroadcastReceiver extends BroadcastReceiver
    {
        @Override
        public void onReceive(Context context, Intent intent)
        {
            if (btState == State.UNAVAILABLE)
            {
                return;
            }
            final String action = intent.getAction();
            Log.d(TAG, "BluetoothHeadsetBroadcastReciever.onReceive: a=" + action);
            
            // Change in connection state of the Headset profile. Note that the
            // change does not tell us anything about whether we're streaming
            // audio to BT over SCO. Typically received when user turns on a BT
            // headset while audio is active using another audio device.
            
            if (action.equals(BluetoothHeadset.ACTION_CONNECTION_STATE_CHANGED))
            {
                final int state = intent.getIntExtra(BluetoothHeadset.EXTRA_STATE, BluetoothHeadset.STATE_DISCONNECTED);
                Log.d(TAG, "BluetoothHeadsetBroadcastReceiver.onReceive: "
                            + "a=ACTION_CONNECTION_STATE_CHANGED, " 
                            + "s=" + stateToString(state) + ", " + "sb="
                            + isInitialStickyBroadcast() + ", " + "BT state: " + state);
                if (state == BluetoothHeadset.STATE_CONNECTED)
                {
                    connectionAttempts = 0;
                    //Device connnected
                    updateAudioDeviceState();
                }
                else if (state == BluetoothHeadset.STATE_CONNECTING)
                {
                    // No action needed.
                }
                else if (state == BluetoothHeadset.STATE_DISCONNECTING)
                {
                    // No action needed.
                }
                else if (state == BluetoothHeadset.STATE_DISCONNECTED)
                {
                    // Bluetooth is probably powered off during the call.
                    stopAudio();
                    //Pick another device?
                    updateAudioDeviceState();
                }
                // Change in the audio (SCO) connection state of the Headset profile.
                // Typically received after call to startScoAudio() has finalized.
            }
            else if (action.equals(BluetoothHeadset.ACTION_AUDIO_STATE_CHANGED))
            {
                final int state = intent.getIntExtra(BluetoothHeadset.EXTRA_STATE,
                                                     BluetoothHeadset.STATE_AUDIO_DISCONNECTED);
                Log.d(TAG, "BluetoothHeadsetBroadcastReceiver.onReceive: " + "a=ACTION_AUDIO_STATE_CHANGED, " 
                            + "s=" + stateToString(state) + ", " + "sb="
                            + isInitialStickyBroadcast() + ", " + "BT state: " + state);
                if (state == BluetoothHeadset.STATE_AUDIO_CONNECTED)
                {
                    cancelTimer();
                    if (btState == State.CONNECTING)
                    {
                        Log.d(TAG, "+++ Bluetooth audio SCO is now connected");
                        btState = State.CONNECTED;
                        connectionAttempts = 0;
                        updateAudioDeviceState();
                    }
                    else
                    {
                        Log.w(TAG, "Unexpected state BluetoothHeadset.STATE_AUDIO_CONNECTED state is:" + btState);
                    }
                }
                else if (state == BluetoothHeadset.STATE_AUDIO_CONNECTING)
                {
                    Log.d(TAG, "+++ Bluetooth audio SCO is now connecting...");
                }
                else if (state == BluetoothHeadset.STATE_AUDIO_DISCONNECTED)
                {
                    Log.d(TAG, "+++ Bluetooth audio SCO is now disconnected");
                    if (isInitialStickyBroadcast())
                    {
                        Log.d(TAG, "Ignore STATE_AUDIO_DISCONNECTED initial sticky broadcast.");
                        return;
                    }
                    updateAudioDeviceState();
                }
            }
            Log.d(TAG, "onReceive done: BT state=" + btState);
        }
    };

    /** 
     * Runs when the Bluetooth timeout expires. We use that timeout after calling
     * startAudio() or stopAudio() because we're not guaranteed to get a
     * callback after those calls. */
    private final Runnable bluetoothTimeoutRunnable = new Runnable()
    {
        @Override
        public void run()
        {
            bluetoothTimeout();
        }
    };
    
    /**
     * Private constructor
     * @param context Application context
     * @param audioDeviceManager AudioDeviceManager instance
     */
    private BluetoothDeviceManagerImpl(final Context context, final AudioDeviceManager audioDeviceManager)
    {
        this.context = context;
        this.btState =  State.UNAVAILABLE;
        this.btServiceListener = new BluetoothServiceListener();
        this.broadcastReceiver = new BluetoothHeadsetBroadcastReceiver();
        this.audioManager = (AudioManager) context.getSystemService(Context.AUDIO_SERVICE);
        this.fcsdkAudioManager = audioDeviceManager;
        handler = new Handler(context.getMainLooper());
    }

    /**
     * Static constructor for BluetoothDeviceManagerImpl
     * @param context Application context
     * @param audioManager AudioDeviceManager
     * @return BluetoothDeviceManagerimpl
     */
    public static BluetoothDeviceManager create(final Context context, final AudioDeviceManager audioManager)
    {
        return new BluetoothDeviceManagerImpl(context, audioManager);
    }
    
    /**
     * Activates components required to detect Bluetooth devices and to enable
     * BT SCO (audio is routed via BT SCO) for the headset profile. The end
     * state will be HEADSET_UNAVAILABLE but a state machine has started which
     * will start a state change sequence where the final outcome depends on
     * if/when the BT headset is enabled.
     * Example of state change sequence when start() is called while BT device
     * is connected and enabled:
     *   UNAVAILABLE --> DEVICE_UNAVAILABLE --> DEVICE_AVAILABLE -->
     *   CONNECTING --> CONNECTED <==> audio is now routed via BT SCO.
     * Note that the AppRTCAudioManager is also involved in driving this state
     * change.
     */
    @Override
    public void start()
    {
        Log.d(TAG, "start");
        if (!hasPermission(context, android.Manifest.permission.BLUETOOTH))
        {
            Log.w(TAG, "Process (pid=" + Process.myPid() + ") lacks BLUETOOTH permission");
            return;
        }
        if (btState != State.UNAVAILABLE)
        {
            Log.w(TAG, "Invalid BT state");
            return;
        }
        btHeadset = null;
        btDevice = null;
        connectionAttempts = 0;
        // Get a handle to the default local Bluetooth adapter.
        btAdapter = BluetoothAdapter.getDefaultAdapter();
        if (btAdapter == null)
        {
            Log.w(TAG, "Device does not support Bluetooth");
            return;
        }
        //Check if the BT adpater is turned on.
        if (!btAdapter.isEnabled())
        {
            Log.w(TAG, "Device has Bluetooth turned off");
            return;
        }
        // Ensure that the device supports use of BT SCO audio for off call use cases.
        if (!audioManager.isBluetoothScoAvailableOffCall())
        {
            Log.e(TAG, "Bluetooth SCO audio is not available off call");
            return;
        }

        logBluetoothAdapterInfo(btAdapter);
        // Establish a connection to the HEADSET profile (includes both Bluetooth Headset and
        // Hands-Free) proxy object and install a listener.
        if (!getBluetoothProfileProxy(context, btServiceListener, BluetoothProfile.HEADSET))
        {
            Log.e(TAG, "BluetoothAdapter.getProfileProxy(HEADSET) failed");
            return;
        }
        
        // Register receivers for BluetoothHeadset change notifications.
        final IntentFilter bluetoothHeadsetFilter = new IntentFilter();
        // Register receiver for change in connection state of the Headset profile.
        bluetoothHeadsetFilter.addAction(BluetoothHeadset.ACTION_CONNECTION_STATE_CHANGED);
        // Register receiver for change in audio connection state of the Headset profile.
        bluetoothHeadsetFilter.addAction(BluetoothHeadset.ACTION_AUDIO_STATE_CHANGED);
        
        registerReceiver(broadcastReceiver, bluetoothHeadsetFilter);
        
        Log.d(TAG, "HEADSET profile state: "
                + stateToString(btAdapter.getProfileConnectionState(BluetoothProfile.HEADSET)));
        Log.d(TAG, "Bluetooth proxy for headset profile has started");
        btState = State.DEVICE_UNAVAILABLE;
        Log.d(TAG, "start done: BT state=" + btState);
    }

    @Override
    public void stop()
    {
        if (!hasPermission(context, android.Manifest.permission.BLUETOOTH))
        {
            Log.w(TAG, "Process (pid=" + Process.myPid() + ") lacks BLUETOOTH permission");
            return;
        }
        unregisterReceiver(broadcastReceiver);
        Log.d(TAG, "stop: BT state=" + btState);
        if (btAdapter != null)
        {
            // Stop BT SCO connection with remote device if needed.
            stopAudio();
            
            // Close down remaining BT resources.
            if (btState != State.UNAVAILABLE)
            {
                cancelTimer();
                if (btHeadset != null)
                {
                    btAdapter.closeProfileProxy(BluetoothProfile.HEADSET, btHeadset);
                    btHeadset = null;
                }
                btAdapter = null;
                btDevice = null;
                btState = State.UNAVAILABLE;
            }
        }
        Log.d(TAG, "stop done: BT state=" + btState);
    }

    @Override
    public boolean deviceIsAvailable()
    {
        return btState == BluetoothDeviceManager.State.DEVICE_AVAILABLE
                || btState == BluetoothDeviceManager.State.DEVICE_UNAVAILABLE
               || btState == BluetoothDeviceManager.State.DISCONNECTING;
    }
    
    @Override
    public boolean deviceIsConnected()
    {
        
        return btState == State.CONNECTED 
                || btState == State.CONNECTING
                || btState == State.DEVICE_AVAILABLE;
    }
    
    @Override
    public State getState()
    {
        return btState;
    }
    
    /** Starts timer which times out after BLUETOOTH_SCO_TIMEOUT_MS milliseconds. */
    private void startTimer()
    {
        Log.d(TAG, "startTimer");
        handler.postDelayed(bluetoothTimeoutRunnable, BLUETOOTH_SCO_TIMEOUT_MS);
    }

    @Override
    public boolean startAudio()
    {
        Log.d(TAG, "startSco: BT state=" + btState + ", "
            + "attempts: " + connectionAttempts + ", "
            + "SCO is on: " + audioManager.isBluetoothScoOn());
        if (connectionAttempts >= 2)
        {
            Log.e(TAG, "BT SCO connection fails - no more attempts");
            return false;
        }
        if (btState != State.DEVICE_AVAILABLE)
        {
            Log.e(TAG, "BT SCO connection fails - no headset available");
            return false;
        }
        // Start BT SCO channel and wait for ACTION_AUDIO_STATE_CHANGED.
        Log.d(TAG, "Starting Bluetooth SCO and waits for ACTION_AUDIO_STATE_CHANGED...");
        // The SCO connection establishment can take several seconds, hence we cannot rely on the
        // connection to be available when the method returns but instead register to receive the
        // intent ACTION_SCO_AUDIO_STATE_UPDATED and wait for the state to be SCO_AUDIO_STATE_CONNECTED.
        
        btState = State.CONNECTING;
        audioManager.setMode(AudioManager.MODE_IN_COMMUNICATION);
        audioManager.setBluetoothScoOn(true);
        audioManager.startBluetoothSco();
        connectionAttempts++;
        startTimer();
    
        Log.d(TAG, "startScoAudio done: BT state=" + btState);
        return true;
    }

    /** Stops Bluetooth SCO connection with remote device. */
    @Override
    public void stopAudio()
    {
        Log.d(TAG, "stopScoAudio: BT state=" + btState + ", " + "SCO is on: " + audioManager.isBluetoothScoOn());
        if (btState != State.CONNECTING && btState != State.CONNECTED)
        {
            return;
        }
        cancelTimer();
        audioManager.stopBluetoothSco();
        audioManager.setBluetoothScoOn(false);
        audioManager.setMode(AudioManager.MODE_NORMAL);
        btState = State.DISCONNECTING;
        Log.d(TAG, "stopScoAudio done: BT state=" + btState);

    }

    protected void registerReceiver(BroadcastReceiver receiver, IntentFilter filter)
    {
        context.registerReceiver(receiver, filter);
    }

    protected void unregisterReceiver(BroadcastReceiver receiver)
    {
        context.unregisterReceiver(receiver);
    }

    protected boolean getBluetoothProfileProxy(Context context, BluetoothProfile.ServiceListener listener, int profile)
    {
        return btAdapter.getProfileProxy(context, listener, profile);
    }

    protected boolean hasPermission(Context context, String permission)
    {
        return context.checkPermission(permission, Process.myPid(), Process.myUid())
           == PackageManager.PERMISSION_GRANTED;
    }

    /** Logs the state of the local Bluetooth adapter. */
    protected void logBluetoothAdapterInfo(BluetoothAdapter localAdapter)
    {
        Log.d(TAG, "BluetoothAdapter: " + "enabled=" + localAdapter.isEnabled() + ", "
                + "state=" + stateToString(localAdapter.getState()) + ", " + "name="
                + localAdapter.getName() + ", " + "address=" + localAdapter.getAddress());
        // Log the set of BluetoothDevice objects that are bonded (paired) to the local adapter.
        final Set<BluetoothDevice> pairedDevices = localAdapter.getBondedDevices();
        if (!pairedDevices.isEmpty())
        {
            Log.d(TAG, "paired devices:");
            for (BluetoothDevice device : pairedDevices)
            {
                Log.d(TAG, " name=" + device.getName() + ", address=" + device.getAddress());
            }
        }
    }

    /** Ensures that the audio manager updates its list of available audio devices. */
    private void updateAudioDeviceState()
    {
        Log.d(TAG, "updateAudioDeviceState");
        fcsdkAudioManager.updateAudioDeviceState();
    }

    /**
     * Use the BluetoothHeadset proxy object (controls the Bluetooth Headset Service via IPC) to update the list of
     * connected devices for the HEADSET profile. The internal state will change to DEVICE_UNAVAILABLE or to
     * DEVICE_AVAILABLE and |bluetoothDevice| will be mapped to the connected device if available.
     */
    @Override
    public void refreshListOfDevices()
    {
        if (btState == State.UNAVAILABLE || btHeadset == null)
        {
            return;
        }
        Log.d(TAG, "updateDevice");
        // Get connected devices for the headset profile. Returns the set of
        // devices which are in state STATE_CONNECTED. The BluetoothDevice class
        // is just a thin wrapper for a Bluetooth hardware address.
        final List<BluetoothDevice> devices = btHeadset.getConnectedDevices();
        if (devices.isEmpty())
        {
            btDevice = null;
            btState = State.DEVICE_UNAVAILABLE;
            Log.d(TAG, "No connected bluetooth headset");
        }
        else
        {
            // Always use first device is list. Android only supports one device.
            btDevice = devices.get(0);
            btState = State.DEVICE_AVAILABLE;
            Log.d(TAG,
                  "Connected bluetooth headset: " + "name=" + btDevice.getName() + ", " + "state="
                       + stateToString(btHeadset.getConnectionState(btDevice)) + ", SCO audio="
                       + btHeadset.isAudioConnected(btDevice));
        }
        Log.d(TAG, "updateDevice done: BT state=" + btState);
    }

    /**
     * Called when start of the BT SCO channel takes too long time. Usually happens when the BT device has been turned
     * on during an ongoing call.
     */
    private void bluetoothTimeout()
    {
        if (btState == State.UNAVAILABLE || btHeadset == null)
        {
            return;
        }
        Log.d(TAG, "bluetoothTimeout: BT state=" + btState + ", "
                + "attempts: " + connectionAttempts + ", "
                + "SCO is on: " + audioManager.isBluetoothScoOn());
        if (btState != State.CONNECTING)
        {
            return;
        }
        // Bluetooth SCO should be connecting; check the latest result.
        boolean scoConnected = false;
        final List<BluetoothDevice> devices = btHeadset.getConnectedDevices();
        if (devices.size() > 0)
        {
            btDevice = devices.get(0);
            if (btHeadset.isAudioConnected(btDevice))
            {
                Log.d(TAG, "SCO connected with " + btDevice.getName());
                scoConnected = true;
            }
            else
            {
                Log.d(TAG, "SCO is not connected with " + btDevice.getName() 
                    + " state " + stateToString(btHeadset.getConnectionState(btDevice)));
            }
        }
        if (scoConnected)
        {
            // We thought BT had timed out, but it's actually on; updating state.
            btState = State.CONNECTED;
            connectionAttempts = 0;
        }
        else
        {
            // Give up and "cancel" our request by calling stopBluetoothSco().
            Log.w(TAG, "BT failed to connect after timeout");
            stopAudio();
        }
        updateAudioDeviceState();
        Log.d(TAG, "bluetoothTimeout done: BT state=" + btState);
    }

    /** Cancels any outstanding timer tasks. */
    private void cancelTimer()
    {
        Log.d(TAG, "cancelTimer");
        handler.removeCallbacks(bluetoothTimeoutRunnable);
    }

    /** Converts BluetoothAdapter states into local string representations. */
    private String stateToString(int state)
    {
        switch (state)
        {
            case BluetoothAdapter.STATE_DISCONNECTED:
                return "DISCONNECTED";
            case BluetoothAdapter.STATE_CONNECTED:
                return "CONNECTED";
            case BluetoothAdapter.STATE_CONNECTING:
                return "CONNECTING";
            case BluetoothAdapter.STATE_DISCONNECTING:
                return "DISCONNECTING";
            case BluetoothAdapter.STATE_OFF:
                return "OFF";
            case BluetoothAdapter.STATE_ON:
                return "ON";
            case BluetoothAdapter.STATE_TURNING_OFF:
                // Indicates the local Bluetooth adapter is turning off. Local clients should immediately
                // attempt graceful disconnection of any remote links.
                return "TURNING_OFF";
            case BluetoothAdapter.STATE_TURNING_ON:
                // Indicates the local Bluetooth adapter is turning on. However local clients should wait
                // for STATE_ON before attempting to use the adapter.
                return "TURNING_ON";
            default:
                return "INVALID";
        }
    }

}
