/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import org.acbrtc.StatsReport;

import android.util.Log;

/*
 * Helper class to convert raw call statistics into a quality indicator as discussed in WEBRTC-5322.
 */
class CallQualityCalculator
{
	protected static final String TAG = "CallQualityCalculator";
	
	private int previousAudioPacketsReceived;
	private int previousAudioPacketsLost;
	private int previousVideoPacketsReceived;
	private int previousVideoPacketsLost;
	private int inboundQuality;
	
	public CallQualityCalculator()
	{
		
	}
	
	/**
     * We use the R-factor Function value for G711 as the quality indicator.
	 * See http://www.voiptroubleshooter.com/diagnosis/emodel.html.
	 * Also see discussion in WEBRTC-5322 for further information.
	 */
	public int calculateQuality(StatsReport[] remoteAudioStats, StatsReport[] remoteVideoStats)
	{
		int percentageAudioPacketsReceived = 0;
		int percentageVideoPacketsReceived = 0;		
		int averagePercentage = 0;
		
		if ((remoteAudioStats != null) && (remoteVideoStats != null))
		{
			percentageAudioPacketsReceived = calculatePercentageAudioPacketsReceived(remoteAudioStats);
			percentageVideoPacketsReceived = calculatePercentageVideoPacketsReceived(remoteVideoStats);

			if ((percentageAudioPacketsReceived == 0) || (percentageVideoPacketsReceived == 0))
			{
				// Don't let zero audio/video packets affect average.
				averagePercentage = percentageAudioPacketsReceived + percentageVideoPacketsReceived;
			}
			else
			{
				averagePercentage = (percentageAudioPacketsReceived + percentageVideoPacketsReceived) / 2;
			}
		}
		else if (remoteAudioStats != null)
		{
			percentageAudioPacketsReceived = calculatePercentageAudioPacketsReceived(remoteAudioStats);
			averagePercentage = percentageAudioPacketsReceived;
		}
		else if (remoteVideoStats != null)
		{
			percentageVideoPacketsReceived = calculatePercentageVideoPacketsReceived(remoteVideoStats);
			averagePercentage = percentageVideoPacketsReceived;
		}
		
		try
		{
			int loss = 100 - averagePercentage;
			inboundQuality = 100 - (95 * loss / (loss + 10));
		}
	    catch (ArithmeticException e) 
		{
	    	Log.d(TAG, "CallQualityCalculator quality calculation failed averagePercentage=" + averagePercentage + " audio=" + remoteAudioStats + " video=" + remoteVideoStats);
	    	inboundQuality = -1;
	    }
		
		Log.d(TAG, "CallQualityCalculator audio packets received=" + percentageAudioPacketsReceived +
		                                " video packets received=" + percentageVideoPacketsReceived +
				                        " quality=" + inboundQuality);
		
		return inboundQuality;
	}
	
	/**
	 */
	private int calculatePercentageAudioPacketsReceived(StatsReport[] statsReport) 
	{
		int totalAudioPacketsReceived = getValueFromStatsReportId(statsReport, "ssrc", "packetsReceived");
		int totalAudioPacketsLost = getValueFromStatsReportId(statsReport, "ssrc", "packetsLost");
		
		int newAudioPacketsReceived = totalAudioPacketsReceived - previousAudioPacketsReceived;
		int newAudioPacketsLost = totalAudioPacketsLost - previousAudioPacketsLost;
		
		Log.d(TAG, "CQ stats - previous audio packets=" + previousAudioPacketsReceived 
				+ ", previous audio packets lost=" + previousAudioPacketsLost);
		
		previousAudioPacketsReceived = totalAudioPacketsReceived;
		previousAudioPacketsLost = totalAudioPacketsLost;

		Log.d(TAG, "CQ stats - total audio packets=" + totalAudioPacketsReceived 
				+ ", total audio packets lost=" + totalAudioPacketsLost);
		Log.d(TAG, "CQ stats - new audio packets=" + newAudioPacketsReceived 
				+ ", new audio packets lost=" + newAudioPacketsLost);
		
		return calculatePacketReceivedPercentage(newAudioPacketsReceived,
                newAudioPacketsLost);
	}
	
	/**
	 */
	private int calculatePercentageVideoPacketsReceived(StatsReport[] statsReport) 
	{
		int totalVideoPacketsReceived = getValueFromStatsReportId(statsReport, "ssrc", "packetsReceived");
		int totalVideoPacketsLost = getValueFromStatsReportId(statsReport, "ssrc", "packetsLost");

		Log.d(TAG, "CQ stats - previous video packets=" + previousVideoPacketsReceived 
				+ ", previous video packets lost=" + previousVideoPacketsLost);
		
		// hold and mute video causes PeerConnection to reset the total packets received to 0
		// so must reset our previous values to 0. This does not happen for audio packets
		if (packetsReceivedHasBeenReset(totalVideoPacketsReceived, previousVideoPacketsReceived))
		{
			previousVideoPacketsReceived = 0;
		}
		if (packetsLostHasBeenReset(totalVideoPacketsLost, previousVideoPacketsLost))
		{
			previousVideoPacketsLost = 0;
		}
		
		int newVideoPacketsReceived = totalVideoPacketsReceived - previousVideoPacketsReceived;
		int newVideoPacketsLost = totalVideoPacketsLost - previousVideoPacketsLost;

		previousVideoPacketsReceived = totalVideoPacketsReceived;
		previousVideoPacketsLost = totalVideoPacketsLost;
		
		Log.d(TAG, "CQ stats - total video packets=" + totalVideoPacketsReceived 
				+ ", total video packets lost=" + totalVideoPacketsLost);
		Log.d(TAG, "CQ stats - new video packets=" + newVideoPacketsReceived 
				+ ", new video packets lost=" + newVideoPacketsLost);
			
		return calculatePacketReceivedPercentage(newVideoPacketsReceived, newVideoPacketsLost);		
	}
		
	private boolean packetsLostHasBeenReset(int packetsLost, int previousPacketLost)
    {
	    return packetsLost < previousPacketLost;
    }

	private boolean packetsReceivedHasBeenReset(int packetsLost, int previousPacketLost)
    {
	    return packetsLost < previousPacketLost;
    }

	private int calculatePacketReceivedPercentage(int newPacketsReceived,
            int newPacketsLost)
    {
	    // defend against arithmetic exception as I believe peer connection stats can be negative
		// in some circumstances
		if (newPacketsLost < 0 || newPacketsReceived < 0)
		{
			return 0;
		}
		
		// Calculate percentage received since last sample.
		if (newPacketsLost == 0)
		{
			if (newPacketsReceived == 0)
			{
				return 0;
			}
			else
			{
				return 100;
			}
		}
		else
		{
			return (newPacketsReceived * 100) / (newPacketsReceived + newPacketsLost);
		}
    }

	/**
	 * 
	 * @param id e.g. ssrc_4163537872 below. 
	 * @param valueName e.g. packetsReceived, googFrameRateDecoded etc.
	 * @return value of the StatsReport.Value specified by valueName
	 * 
	 * Video Track statistics id: ssrc_4163537872, 
	 * type: ssrc, 
	 * timestamp: 1.362355743499847E12, 
	 * values: 	[ssrc: 4163537872], [googTrackId: dXmHEefcqcm5IQmqcl3IFm0r5H69oOZ4fYYRv0], 
	 * 			[transportId: Channel-audio-1], [bytesReceived: 4132493], [packetsReceived: 4748], 
	 * 			[packetsLost: 677], [googFirsSent: 0], [googNacksSent: -1], [googFrameWidthReceived: 640], 
	 * 			[googFrameHeightReceived: 480], [googFrameRateReceived: 15], [googFrameRateDecoded: 14], 
	 * 			[googFrameRateOutput: 14], 
	 */
	private int getValueFromStatsReportId(StatsReport[] statsReport, String id, String valueName) 
	{
		// TODO: Handle non-integer values.

		for (StatsReport s : statsReport)
		{
			if (s.id.contains(id)) // e.g. ssrc_4163537872
			{
				for (StatsReport.Value v : s.values) // [ssrc: 4163537872] ... [googFrameRateOutput: 14]
				{
					if (v.name.equalsIgnoreCase(valueName)) // e.g. packetsReceived
					{
						return Integer.valueOf(v.value).intValue(); // e.g. 4748
					}
				}
			}
		}

		return 0;
	}
}
