/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.util.Hashtable;
import java.util.Timer;
import java.util.TimerTask;

import org.acbrtc.AudioTrack;
import org.acbrtc.MediaStream;
import org.acbrtc.PeerConnection;
import org.acbrtc.StatsObserver;
import org.acbrtc.StatsReport;
import org.acbrtc.VideoTrack;

import android.content.Context;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;

/**
 * Periodically obtain call statistics from WebRTC.
 * 
 * Only remote stream statistics are obtained ATM.
 */
public class CallStatisticsMonitor 
{
	private static final String	TAG = "CallStatisticsMonitor";
	
    private final static long ONE_SECOND = 1000;
    private final static int FIVE_SECONDS = 5000;

    private static final String	MEDIA_STREAM_LABEL_REMOTE	= "remote_media_stream";

	private PeerConnection mPeerConnection;
	private Hashtable<String, MediaStream>	mActiveStreams;
	private CallStatisticsListener mListener;
	Context mContext;
	private boolean mRunning;

	private TimerTask mTimerTask;
	private Timer mTimer;
	
	private StatsReport[] mLastFiveSecondsRemoteAudioStats;
	private StatsReport[] mLastFiveSecondsRemoteVideoStats;

	private Object csmLock = new Object();

	
	/**
	 */
	CallStatisticsMonitor(PeerConnection peerConnection, Hashtable<String, MediaStream> activeStreams, CallStatisticsListener listener, Context context)
	{
		if ((peerConnection == null) || (activeStreams == null) || (listener == null) || (context == null))
		{
			throw new IllegalArgumentException("Constructor arguments cannot be null");
		}
		
		synchronized(csmLock)
		{
			mPeerConnection = peerConnection;
			mActiveStreams = activeStreams;
			mListener = listener;
			mContext = context;
			mRunning = false;
		}
	}

	/**
	 */
	public void start()
	{
		synchronized(csmLock)
		{
		    Log.i(TAG, "Starting CallStatisticsMonitor");
	
		    // The TimerTask will continue for the duration of the call, regardless of
		    // whether the call is audio or video-only, held, muted, or the application
		    // is put into the background or rotated.
	
			mTimer = new Timer();
		    mTimerTask = new GetStatisticsTimerTask();
	
		    mTimer.scheduleAtFixedRate(mTimerTask, FIVE_SECONDS, FIVE_SECONDS);
		    
		    mRunning = true;
		    
		    Log.i(TAG, "Started CallStatisticsMonitor");
		}
	}

	/**
	 */
	public void stop()
	{
		synchronized(csmLock)
		{
			Log.i(TAG, "Stopping CallStatisticsMonitor");
	
			mRunning = false;
			
			mLastFiveSecondsRemoteAudioStats = null;
			mLastFiveSecondsRemoteVideoStats = null;
			
			if (mTimerTask != null)
			{
				mTimerTask.cancel();
				mTimerTask = null;
			}
			
			if (mTimer != null)
			{
				mTimer.cancel();
				mTimer.purge();
				mTimer = null;
			}
			
			Log.i(TAG, "Stopped CallStatisticsMonitor");
		}
	}

	/**
	 */
	private class GetStatisticsTimerTask extends TimerTask
	{
		public void run() 
	    {
			synchronized(csmLock)
			{
				if (mRunning)
				{
					// Need to execute the call to PeerConnection.getStats on the main UI thread.
					// See https://jira.cafex.com/jira/browse/WEBRTC-5624.
					new Handler(mContext.getMainLooper()).post(new GetStatisticsRunnable());
					
					// Wait for the above getStats calls to complete as they run/callback on a separate thread. 
		            try 
		            { 
		               Thread.sleep(ONE_SECOND );
		            } 
		            catch (InterruptedException e) {}
		    
		            // Pass on statistics even if they're null.
	                if ((mRunning) && (mListener != null))
	                {
	                    mListener.onCallStatisticsData(mLastFiveSecondsRemoteAudioStats, mLastFiveSecondsRemoteVideoStats);
	                }
				}
			}
	    }
	}
	
	/**
	 */
	private class GetStatisticsRunnable implements Runnable
	{
		public void run() 
	    {
			if (!mRunning)
			{
				Log.i(TAG, "Not collecting remote stream call statistics");
				return;
			}
			
			Log.i(TAG, "Collecting remote stream call statistics");

			if (mActiveStreams == null)
			{
			    Log.i(TAG, "Not collecting remote stream call statistics as there are no active streams");
			    return;
			}
			
	        MediaStream remoteStreams = mActiveStreams.get(MEDIA_STREAM_LABEL_REMOTE);

			if (remoteStreams == null)
			{
			    Log.i(TAG, "Not collecting remote stream call statistics as there are no remote streams");
			    return;
			}
	        
			AudioTrack remoteAudioTrack = remoteStreams.audioTracks.getFirst();
			
			if (remoteAudioTrack != null)
			{
		    	mPeerConnection.getStats(new StatsObserver() 
		    	    {
		    	        public void onComplete(StatsReport[] stats) 
		    		    {
		    	        	mLastFiveSecondsRemoteAudioStats = stats;
		    		    }				
					}, remoteAudioTrack);
			}
	
			VideoTrack remoteVideoTrack = remoteStreams.videoTracks.getFirst();
	
			if (remoteVideoTrack != null)
			{
	            mPeerConnection.getStats(new StatsObserver() 
	                {
					        public void onComplete(StatsReport[] stats) 
						    {
					        	mLastFiveSecondsRemoteVideoStats = stats;
						    }
						} , remoteVideoTrack);
			}
			
			Log.i(TAG, "Collected remote stream call statistics - audio= " + remoteAudioTrack + " video= " + remoteVideoTrack);
	    }
	}
}
	