/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import android.hardware.Camera;
import android.hardware.Camera.CameraInfo;
import android.util.Log;

import com.alicecallsbob.fcsdk.android.phone.PhoneVideoCaptureResolution;
import com.alicecallsbob.fcsdk.android.phone.PhoneVideoCaptureSetting;
import com.alicecallsbob.fcsdk.android.phone.ResolutionNotFoundException;

/**
 * This offers analysis of the device Cameras. It allows the client to obtain
 * details such as the supported camera resolutions and supported framerates for
 * each camera.
 */
public class CameraVideoAnalyser
{
	/** TAG */ 
	private static final String TAG = CameraVideoAnalyser.class.getSimpleName();

	/** The height of a 720p HD resolution */
	private static final int SEVEN_TWENTY_RESOLUTION_HEIGHT = 720;

	/** The minimum number of cores required to support 720p or higher */
	private static final int MIN_CORES_FOR_720 = 2;

	protected static final int ANDROID_FRAMERATE_MULTIPLIER = 1000;

	protected static final int SUGGESTED_FRAMERATE = 30;

	private Map<Integer, List<int[]>> supportedCameraFramerates;

	private Map<Integer, List<PhoneVideoCaptureSetting>> supportedCameraSettings;

	public CameraVideoAnalyser()
	{
		supportedCameraFramerates = new HashMap<Integer, List<int[]>>();
		supportedCameraSettings = new HashMap<Integer, List<PhoneVideoCaptureSetting>>();
	}

	/**
	 * Gets a Map containing the supported camera framerates for each camera on
	 * the device. The List of framerates are keyed by the camera index Id.
	 * 
	 * In order for this to be populated setSupportedCamerasSettingsAndFramerates()
	 * should be called first.
	 * 
	 * @return Map containing supported camera framerates
	 */
	public Map<Integer, List<int[]>> getSupportedCameraFramerates()
	{
		return supportedCameraFramerates;
	}

	/**
	 * Gets a Map containing the supported camera settings for each camera on
	 * the device. The List of settings are keyed by the camera index Id.
	 * 
	 * In order for this to be populated setSupportedCamerasSettingsAndFramerates()
	 * should be called first.
	 * 
	 * @return Map containing supported camera settings
	 */
	public Map<Integer, List<PhoneVideoCaptureSetting>> getSupportedCameraSettings()
	{
		return supportedCameraSettings;
	}

	/**
	 * Gets the default camera ID for the device.  It will default to the front camera if there is one.
	 * 
	 * @return the default camera ID
	 */
	public int getDefaultCameraIdxToUse()
	{
		int defaultCameraIdx = 0;

		final int numberOfCameras = Camera.getNumberOfCameras();
		if (numberOfCameras == 1)
		{
			Log.d(TAG,
			        "Only 1 camera on device, SDK defaulting to camera with idx 0");
		}
		else
		{
			Log.d(TAG,
			        "More than 1 camera on device, SDK defaulting to the first FRONT camera we can find...");

			boolean cameraFound = false;
			for (int i = 0; i < numberOfCameras; i++)
			{
				final CameraInfo cameraInfo = new CameraInfo();
				Camera.getCameraInfo(i, cameraInfo);

				if (cameraInfo.facing == CameraInfo.CAMERA_FACING_FRONT)
				{
					defaultCameraIdx = i;
					Log.d(TAG, "Using FRONT facing camera at idx: " + i);
					cameraFound = true;
					break;
				}
			}

			if (!cameraFound)
			{
				Log.w(TAG,
				        "SDK failed to find a FRONT facing camera - defaulting to camera at idx 0");
			}
		}
		return defaultCameraIdx;
	}	

	/**
	 * Sets the supported settings and framerates for each camera.
	 * 
	 * @param supportedCameraSettings
	 * @param supportedCameraFramerates
	 */
	public void setSupportedCamerasSettingsAndFramerates()
	{
		final int numberOfCameras = Camera.getNumberOfCameras();
		Log.d(TAG, "Number of cameras on device: " + numberOfCameras);

		Camera camera = null;
		for (int i = 0; i < numberOfCameras; i++)
		{
			// add the cameraindex to a map for later checking
			supportedCameraSettings.put(i,
			        new ArrayList<PhoneVideoCaptureSetting>());
			supportedCameraFramerates.put(i, new ArrayList<int[]>());

			try
			{
				camera = Camera.open(i);
				setCameraResolutionsAndFramerates(camera, i);
			}
			catch (Exception e)
			{
				Log.e(TAG, "Failed to open camera idx [" + i
				        + "] and getParameters()", e);
			}
			finally
			{
				if (camera != null)
				{
					camera.release();
				}
			}
		}
	}

	/**
	 * Checks if the camera with the given cameraId is available to use.
	 * 
	 * @param cameraId
	 * @return true if camera is avaialble, false if not.
	 */
	public boolean isCameraAvailable(int cameraId)
	{
		Camera camera = null;
		try
		{
			camera = Camera.open(cameraId);
			if (cameraSettingsHaveBeenSet(cameraId))
			{
				Log.d(TAG, "camera settings already set");
				return true;
			}
			else
			{
				Log.d(TAG, "Setting camera settings");
				setCameraResolutionsAndFramerates(camera, cameraId);
			}
		}
		catch (Exception e)
		{
			Log.e(TAG, "Failed to open camera idx [" + cameraId
			        + "] and getParameters()", e);
			return false;
		}
		finally
		{
			if (camera != null)
			{
				camera.release();
			}
		}
		return true;
	}

	private boolean cameraSettingsHaveBeenSet(int cameraIndex)
	{
		final List<int[]> camFramerates = supportedCameraFramerates
		        .get(cameraIndex);
		final List<PhoneVideoCaptureSetting> camSettings = supportedCameraSettings
		        .get(cameraIndex);

		if (camFramerates == null || camSettings == null)
		{
			return false;
		}

		return camFramerates.size() > 0 && camSettings.size() > 0;
	}

	private void setCameraResolutionsAndFramerates(Camera camera, int cameraIndex)
	{
		final List<int[]> framerates = camera.getParameters()
		        .getSupportedPreviewFpsRange();
		supportedCameraFramerates.put(cameraIndex, framerates);

		final int framerate = getRecommendedFramerate(framerates);

		final List<PhoneVideoCaptureSetting> settings = getCameraDetails(
		        camera, framerate);
		supportedCameraSettings.put(cameraIndex, settings);
	}
	
	private int getRecommendedFramerate(List<int[]> supportedFramerateRanges)
	{
		return getClosestSupportedFramerate(SUGGESTED_FRAMERATE,
		        supportedFramerateRanges);
	}
	
	/**
	 * Calculates the closest framerate to a specific framerate, from a list of
	 * supported ranges.
	 * 
	 * @param preferredFrameRate
	 *            The framerate
	 * @param supportedFrameRateRanges
	 *            The supported framerate ranges
	 * @return The best supported framerate
	 */
	public int getClosestSupportedFramerate(int preferredFrameRate,
	        List<int[]> supportedFrameRateRanges)
	{
		int preferredFramerateMultiplied = preferredFrameRate
		        * ANDROID_FRAMERATE_MULTIPLIER;
		int framerate = -1;
		int currentDistanceFromSuggested = -1;

		for (int[] framerateRange : supportedFrameRateRanges)
		{
			// The suggested framerate is within the range, so use that
			if (framerateRange[0] <= preferredFramerateMultiplied
			        && framerateRange[1] >= preferredFramerateMultiplied)
			{
				framerate = preferredFrameRate;
				break;
			}
			// The range is higher than the suggested, use the lowest, if its
			// the closest we've found so far
			else if (framerateRange[0] > preferredFramerateMultiplied)
			{
				final int thisDistance = framerateRange[0]
				        - preferredFramerateMultiplied;
				if (currentDistanceFromSuggested == -1
				        || thisDistance < currentDistanceFromSuggested)
				{
					framerate = framerateRange[0]
					        / ANDROID_FRAMERATE_MULTIPLIER;
					currentDistanceFromSuggested = thisDistance;
				}
			}
			// The range is lower than the suggested, use the highest, if its
			// the closest we've found so far
			else
			{
				final int thisDistance = preferredFramerateMultiplied
				        - framerateRange[1];
				if (currentDistanceFromSuggested == -1
				        || thisDistance < currentDistanceFromSuggested)
				{
					framerate = framerateRange[1]
					        / ANDROID_FRAMERATE_MULTIPLIER;
					currentDistanceFromSuggested = thisDistance;
				}
			}
		}

		return framerate;
	}
	
	private List<PhoneVideoCaptureSetting> getCameraDetails(Camera camera,
	        int framerate)
	{
		List<Camera.Size> sizes = null;
		sizes = camera.getParameters().getSupportedPreviewSizes();

		final List<PhoneVideoCaptureSetting> settings = new ArrayList<PhoneVideoCaptureSetting>();

		if (sizes != null)
		{
			for (Camera.Size size : sizes)
			{
				try
				{
					final PhoneVideoCaptureResolution resolution = PhoneVideoCaptureResolution
					        .getFromDimensions(size.width, size.height);

					// If 720p or higher, there must be at least 2 cores
					if (resolution.getHeight() < SEVEN_TWENTY_RESOLUTION_HEIGHT
					        || getNumCores() >= MIN_CORES_FOR_720)
					{
						settings.add(new PhoneVideoCaptureSettingImpl(
						        resolution, framerate));
					}
				}
				catch (ResolutionNotFoundException e)
				{
					// Do nothing...this phone resolution is not supported by
					// FCSDK
				}
			}
		}
		return settings;
	}
	

	/**
	 * Gets the number of cores available in this device, across all processors.
	 * 
	 * @return The number of cores, or 1 if failed to get result
	 */
	private int getNumCores()
	{
		// Private Class to display only CPU devices in the directory listing
		class CpuFilter implements FileFilter
		{
			@Override
			public boolean accept(File pathname)
			{
				// Check if filename is "cpu", followed by a single digit number
				if (Pattern.matches("cpu[0-9]+", pathname.getName()))
				{
					return true;
				}
				return false;
			}
		}

		try
		{
			// Get directory containing CPU info
			File dir = new File("/sys/devices/system/cpu/");
			// Filter to only list the devices we care about
			File[] files = dir.listFiles(new CpuFilter());
			// Return the number of cores (virtual CPU devices)
			return files.length;
		}
		catch (Exception e)
		{
			// Default to return 1 core
			return 1;
		}
	}
}
