/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.util.HashMap;
import java.util.Map;

import org.acbrtc.AudioTrack;
import org.acbrtc.MediaConstraints;
import org.acbrtc.MediaConstraints.KeyValuePair;
import org.acbrtc.MediaStream;
import org.acbrtc.PeerConnectionFactory;
import org.acbrtc.VideoCapturer;
import org.acbrtc.VideoRenderer;
import org.acbrtc.VideoSource;
import org.acbrtc.VideoTrack;

import android.hardware.Camera;
import android.hardware.Camera.CameraInfo;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;

import com.alicecallsbob.fcsdk.android.phone.PhoneListener;
import com.alicecallsbob.fcsdk.android.phone.PhoneVideoCaptureResolution;
import com.alicecallsbob.fcsdk.android.phone.PhoneVideoCaptureSetting;
import com.alicecallsbob.fcsdk.android.phone.VideoSurface;

/**
 * Manages the local media stream and video capturer
 * 
 * @author CafeX Communications
 *
 */
public class LocalMediaManager 
{
    private static final String VIDEO_TRACK = "video_track_label";
    
    /** Audio track label */
    private static final String AUDIO_TRACK = "audio_track_label";
    
    /** Local media stream label */
    private static final String MEDIA_STREAM_LABEL_LOCAL    = "local_media_stream";
    
    /** Tag for logging */
    private static final String TAG = "LocalMediaManager";
    
    /** Instance of this class */
    private static LocalMediaManager INSTANCE;
 
    /** The local video capturer */
    private VideoCapturer mVideoCapturer;
    
    /** The local media stream */
    private MediaStream mLocalStream;
    
    /** The local video preview surface */
    private VideoSurfaceImpl mLocalVideoSurface;
    
    /** The camera index that is to be used */
    private int mCameraIndex;
    
    /** Local video source */
    private VideoSource mVideoSource;
    
    /** The local video orientation */
    private int mOrientation = 0;
    
    /** The local video renderer */
    private VideoRenderer mLocalVideoRenderer;

    private VideoSurfaceImpl.VideoRendererCallbacks localCallbacks;
    
    /** 
     * Private singleton constructor
     */
    private LocalMediaManager()
    {
        
    }
    
    /**
     * Obtains the singleton instance of this class.
     * 
     * @return The instance of this class
     */
    public static LocalMediaManager getInstance()
    {
        if (INSTANCE == null)
        {
            INSTANCE = new LocalMediaManager();
        }
        
        return INSTANCE;
    }
    
    /**
     * Dispose of the video capturer, along with the video source.
     */
    public void disposeVideoCapturer()
    {   
        mLocalVideoRenderer = null;
        
        if (mVideoSource != null)
        {
            mVideoSource.dispose();
            mVideoSource = null;
        }
        
        if (mVideoCapturer != null)
        {
            mVideoCapturer.dispose();
        }                
        
        mVideoCapturer = null;      
        mLocalStream = null;
        mLocalVideoSurface = null;
    }
    
    /**
     * Gets the current video capturer, or creates one if it
     * does not currently exist
     * @param cameraFacingId The camera id to use
     * @return the currently or freshly created video capturer, or null if attempts to create it failed
     */
    private VideoCapturer getOrCreateVideoCapturer(int cameraFacingId)
    {   
        if (mVideoCapturer == null)
        {
            mCameraIndex = cameraFacingId;
            Log.v(TAG, "Creating local video source. Camera ("
                    + (cameraFacingId == CameraInfo.CAMERA_FACING_BACK ? "BACK" : "FRONT"));
            //PeerConnectionFactory.initializeAndroidGlobals(context);
            final String cameraName = getCameraDeviceName(cameraFacingId);
            mVideoCapturer = VideoCapturer.create(cameraName);
            
            if ( mVideoCapturer == null ) // we failed to find the specified Camera, so use the default
            {
                mVideoCapturer = VideoCapturer.create("");
            }
            
            if(mVideoCapturer == null) {
            	// We couldn't find the default camera either, give up
            	return null;
            }
            
            mVideoCapturer.setRotation(mOrientation);            
        }
        
        return mVideoCapturer;
    }
    
    /**
     * Gets the current local stream, or creates one with video if it does
     * not exist.
     * @param factory Peer connection factory
     * @param cameraIndex The camera index to use.
     * @param videoRotation The rotation of the video that is to be captured
     * @param callManager The call manager
     * @return The freshly created local video stream or the current stream.  NOTE: if a local audio only
     * stream is already active, then this will be returned from this method,
     * not a video stream.
     */
    public MediaStream getOrCreateLocalVideoStream(PeerConnectionFactory factory, int cameraIndex, PhoneImpl callManager)
    {
        if (mLocalStream == null)
        {
            getOrCreateLocalAudioStream(factory);
            
            // WEBRTC-6460 - patching this until we refactor for the in-call camera changing fix
            // mVideoCapturer = getOrCreateVideoCapturer(cameraIndex);
            mVideoCapturer = getOrCreateVideoCapturer(callManager.getCamera());            
            
            setupLocalVideo(factory, callManager);
        }
        
        return mLocalStream;
    }
    
    /**
     * Gets the current local stream, or creates one with audio only, if a 
     * stream does not currently exist.
     * @param factory Peer connection factory
     * @return The freshly created local audio stream or the current stream.  NOTE: if a local video
     * stream is already active, then this will be returned from this method,
     * not an audio only stream.
     */
    public MediaStream getOrCreateLocalAudioStream(PeerConnectionFactory factory)
    {
        if (mLocalStream == null)
        {
            /*
             * Add the media streams
             */
            mLocalStream = factory.createLocalMediaStream(MEDIA_STREAM_LABEL_LOCAL);
            mLocalStream.addTrack(factory.createAudioTrack(AUDIO_TRACK));
        }
        
        return mLocalStream;
    }
    
    /**
     * Obtains the device name, based on the camera index and camera info
     * @return The device name
     */
    private static String getCameraDeviceName(final int cameraIndex, final CameraInfo cameraInfo)
    {
        String facingString = "front";
        
        if (cameraInfo.facing == CameraInfo.CAMERA_FACING_BACK)
        {
            facingString = "back";
        }

        final StringBuilder builder = new StringBuilder("Camera ");
        builder.append(cameraIndex);
        builder.append(", Facing ");
        builder.append(facingString);
        builder.append(", Orientation ");
        builder.append(cameraInfo.orientation);
        
        return builder.toString();
    }
    
    /**
     * Obtains the device name, based on which way the camera is facing (front or back)
     * @param cameraFacingID The front or back camera
     * @return The device name
     */
    private static String getCameraDeviceName(final int cameraFacingID)
    {
        for (int i = 0; i < Camera.getNumberOfCameras(); i++)
        {
            CameraInfo cameraInfo = new CameraInfo();
            
            Camera.getCameraInfo(i, cameraInfo);
            
            if (cameraInfo.facing == cameraFacingID)
            {
                return getCameraDeviceName(i, cameraInfo);
            }
        }

        return "";
    }
    
    /**
     * Set up the video components for the local end of the call, e.g. the camera and the
     * SurfaceView.
     * @param stream
     */
    public void setupLocalVideo(PeerConnectionFactory factory, PhoneImpl callManager)
    {
        Log.d(TAG, "setupLocalVideo " + mLocalStream);

        if ( mVideoCapturer != null )
        {
        	if (mVideoSource == null)
        	{
                mVideoSource = factory.createVideoSource(mVideoCapturer, getVideoSourceMediaConstraints(callManager));
        	}

            VideoTrack videoTrack = factory.createVideoTrack(VIDEO_TRACK, mVideoSource);

            if (mLocalStream == null)
            {
                return;
            }
            
            mLocalStream.addTrack(videoTrack);

            addLocalVideoRenderer();
            
            final PhoneVideoCaptureSetting videoSetting = convertVideoFormatToSetting(mVideoCapturer.getCaptureFormat());
            
            for ( PhoneListener listener : callManager.mListeners )
            {
                listener.onCaptureSettingChange(videoSetting, mCameraIndex);
            }
        }
        else
        {
            Log.e(TAG, "Failed to create video capture device");
        }
    }
    
    /**
     * Set the resolution and framerates to be the closest supported values.
     * 
     * @return The media contraints
     */
    private MediaConstraints getVideoSourceMediaConstraints(PhoneImpl callManager)
    {
        final MediaConstraints constraints = new MediaConstraints();
        final PhoneVideoCaptureResolution resolution = callManager.getClosestSupportedResolutionToPreferred();
        constraints.mandatory.add(new KeyValuePair("minWidth", resolution.getWidth().toString()));
        constraints.mandatory.add(new KeyValuePair("maxWidth", resolution.getWidth().toString()));
        constraints.mandatory.add(new KeyValuePair("minHeight", resolution.getHeight().toString()));
        constraints.mandatory.add(new KeyValuePair("maxHeight", resolution.getHeight().toString()));
        constraints.mandatory.add(new KeyValuePair("minFrameRate", callManager.getClosestSupportedFramerateToPreferred().toString()));
        constraints.mandatory.add(new KeyValuePair("maxFrameRate", callManager.getClosestSupportedFramerateToPreferred().toString()));
        
        return constraints;
    }
    
    /**
     * Add the local video view as the renderer for the given MediaStream.
     * @param stream
     */
    private void addLocalVideoRenderer()
    {
        Log.d(TAG, "addLocalVideoRenderer");

        if (mLocalVideoSurface == null)
        {
            Log.d(TAG, "UI local video view is null");
        }
        else if (mLocalStream.videoTracks != null && mLocalStream.videoTracks.size() > 0)
        {
            VideoTrack  vt = mLocalStream.videoTracks.get(0);
            if ( vt != null )
            {
                Log.d(TAG, "Adding LOCAL video renderer " + mLocalStream.videoTracks.size());

                localCallbacks = mLocalVideoSurface.createCallbacks(VideoSurface.Endpoint.LOCAL);
                mLocalVideoRenderer = new VideoRenderer(localCallbacks);
                Log.d(TAG, "*** Adding LOCAL video renderer " + mLocalStream + " lv " + mLocalVideoSurface + " cb " + localCallbacks + " r " + mLocalVideoRenderer + " vt " + vt);

                vt.addRenderer(mLocalVideoRenderer);

            }
            else
            {
                Log.w(TAG, "the video track is null");
            }
        }
    }
    
    private void removeLocalVideoRenderer()
    {
        Log.d(TAG, "removeLocalVideoRenderer");
        
        if (mLocalStream.videoTracks != null && mLocalStream.videoTracks.size() > 0)
        {
            VideoTrack  vt = mLocalStream.videoTracks.get(0);
            if ( vt != null )
            {
                vt.removeRenderer(mLocalVideoRenderer);
                Log.d(TAG, "*** Removed LOCAL video renderer " + mLocalVideoRenderer);

                mLocalVideoRenderer = null;

            }
            else
            {
                Log.w(TAG, "the video track is null");
            }
        }

        // AV-373 - explicitly destroy the callback to break references
        if (localCallbacks != null) {
            localCallbacks.destroy();
        }
    }
    
    private PhoneVideoCaptureSetting convertVideoFormatToSetting(String videoFormat)
    {
        //Video format string is in the format:
        //NV21 640x480x120
        
        //Remove text before space
        final String resolutionAndFramerate = videoFormat.split(" ")[1];
        
        //Split on x
        final String[] splitOnX = resolutionAndFramerate.split("x");
        
        final int width = Integer.parseInt(splitOnX[0]);
        final int height = Integer.parseInt(splitOnX[1]);
        final int framerate = Integer.parseInt(splitOnX[2]);
        
        final PhoneVideoCaptureResolution resolution = 
                PhoneVideoCaptureResolution.getFromDimensions(width, height);
        
        return new PhoneVideoCaptureSettingImpl(resolution, framerate);
    }
    
    /**
    *
    * @param localView
    */
   public void setLocalVideoView(VideoSurface localSurface)
   {
       Log.d(TAG, "setLocalVideoView stream " + mLocalStream);

       if (localSurface == null && mLocalVideoSurface != null) {
           mLocalVideoSurface.destroy();
       }
       
       mLocalVideoSurface = (VideoSurfaceImpl)localSurface;
       
       if ( mLocalStream != null )
       {
           if (localSurface != null)
           {
               addLocalVideoRenderer();
           }
           else
           {
        	   removeLocalVideoRenderer();
           }
       }
   }
   
   public void setVideoOrientation(int orientation)
   {
       mOrientation = orientation;
       
       if (mVideoCapturer != null)
       {
           mVideoCapturer.setRotation(orientation);
       }
   }
   
   public void enableLocalAudio(boolean enable)
   {
       if (mLocalStream != null)
       {
           for ( AudioTrack at : mLocalStream.audioTracks )
           {
               Log.i(TAG, "setting audio track " + at.id() + " as "
                       + (enable ? "enabled" : "disabled"));
               at.setEnabled(enable);
           }
       }
   }

   public void enableLocalVideo(boolean enable)
   {
       if ( mLocalStream != null )
       {
           for ( VideoTrack vt : mLocalStream.videoTracks )
           {
               Log.i(TAG, "setting video track " + vt.id() + " as "
                       + (enable ? "enabled" : "disabled"));
               vt.setEnabled(enable);
           }
       }
   }
   
   /**
    * Retrieves the current enabled state of each track in the local audio
    * media stream, mapped by the track ID. This is used with 
    * {@link #restoreLocalAudioEnabledState(Map)} to save the state before
    * temporarily disabling the local media.
    * 
    * @return Map of enabled state by track ID
    */
   public Map<String, Boolean> getLocalAudioEnabledState()
   {
       HashMap<String, Boolean> states = new HashMap<String, Boolean>();
       if (mLocalStream != null)
       {
           for (AudioTrack at : mLocalStream.audioTracks)
           {
               Log.i(TAG, "getting audio track " + at.id() + " enabled state as " + at.enabled());
               states.put(at.id(), at.enabled()? Boolean.TRUE: Boolean.FALSE);
           }
       }
       return states;
   }
   
   /**
    * Restores the enabled state of each track in the local audio media
    * stream.
    * 
    * @param states  Map of enabled state by track ID
    */
   public void restoreLocalAudioEnabledState(Map<String, Boolean> states)
   {
       if (mLocalStream != null)
       {
           for (AudioTrack at : mLocalStream.audioTracks)
           {
               if (states.containsKey(at.id()))
               {
                   Log.i(TAG, "setting audio track ");
                   at.setEnabled(states.get(at.id()).booleanValue());
               }
           }
       }
   }
   
}
