/*
 * libjingle
 * Copyright 2013, Google Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;

import javax.microedition.khronos.egl.EGL10;
import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.egl.EGLContext;
import javax.microedition.khronos.opengles.GL10;

import org.acbrtc.VideoRenderer;
import org.acbrtc.VideoRenderer.I420Frame;

import com.alicecallsbob.fcsdk.android.phone.VideoSurface;
import com.alicecallsbob.fcsdk.android.phone.VideoSurfaceListener;

import android.content.Context;
import android.graphics.Point;
import android.opengl.GLES20;
import android.opengl.GLSurfaceView;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;
import android.view.View;

/**
 * A GLSurfaceView{,.Renderer} that efficiently renders YUV frames from local &
 * remote VideoTracks using the GPU for CSC.
 */
public class VideoSurfaceImpl extends VideoSurface implements GLSurfaceView.Renderer
{
	private static final String	TAG	= "VideoSurface";
	
	// Stashed setSize operation
	private SetSizeParameters setSizeParams = null;

	protected int width = -1;

	protected int height = -1;

	// Pass-through vertex shader.
	private static final String	VERTEX_SHADER_STRING =
			"varying vec2 interp_tc;\n" +
			"\n" +
			"attribute vec4 in_pos;\n" +
			"attribute vec2 in_tc;\n" +
			"\n" +
			"void main() {\n" +
			"  gl_Position = in_pos;\n" +
			"  interp_tc = in_tc;\n" +
			"}\n";

	// YUV to RGB pixel shader. Loads a pixel from each plane and pass through the matrix.
	private static final String	FRAGMENT_SHADER_STRING =
			"precision mediump float;\n" +
			"varying vec2 interp_tc;\n" +
			"\n" +
			"uniform sampler2D y_tex;\n" +
			"uniform sampler2D u_tex;\n" +
			"uniform sampler2D v_tex;\n" +
			"\n" +
			"void main() {\n" +
			"  float y = texture2D(y_tex, interp_tc).r;\n" +
			"  float u = texture2D(u_tex, interp_tc).r - .5;\n" +
			"  float v = texture2D(v_tex, interp_tc).r - .5;\n" +
			// CSC according to http://www.fourcc.org/fccyvrgb.php
			"  gl_FragColor = vec4(y + 1.403 * v, " +
			"                      y - 0.344 * u - 0.714 * v, " +
			"                      y + 1.77 * u, 1);\n" +
			"}\n";

	// Texture Coordinates mapping the entire texture.
	private static final FloatBuffer	textureCoords =
			directNativeFloatBuffer(new float[]{0, 0, 0, 1, 1, 0, 1, 1});

	private static final FloatBuffer	vertices =
			directNativeFloatBuffer(new float[]{-1, 1, -1, -1, 1, 1, 1, -1});

	/**
	 *
	 */
	static class VideoRendererCallbacks implements VideoRenderer.Callbacks
	{
		private static final String TAG = VideoRendererCallbacks.class.getSimpleName(); 
		
		private VideoSurfaceImpl			mView;
	    private VideoSurfaceImpl.Endpoint	mStream;

		public VideoRendererCallbacks(VideoSurfaceImpl view, Endpoint stream)
		{
			mView = view;
			mStream = stream;
		}

        public void destroy() {
            mView = null;
            mStream = null;
        }

		@Override
		public void renderFrame(I420Frame frame)
		{
			if(mView != null) {
			    mView.queueFrame(frame);
			}
		}

		@Override
		public void setSize(final int width, final int height)
		{
			if ( mView != null )
			{
				mView.queueEvent(new Runnable()
				{
					public void run()
					{
                        if (currentContextExists()) {
                            // remove any queued update
                            mView.setSizeParams = null;
                            mView.setSize(mStream, width, height);
                            mView.width = width;
                            mView.height = height;
                        } else {
                            // no current context. queue the operation
                            mView.setSizeParams = new SetSizeParameters(mStream, width, height);
                        }
					}
				});
			}
		}
	}

    private static boolean currentContextExists() {
        return !((EGL10) EGLContext.getEGL()).eglGetCurrentContext().equals(EGL10.EGL_NO_CONTEXT);
    }

	/**
	 * Inform this View of the dimensions of frames coming from |stream|.
	 * @param stream
	 * @param width
	 * @param height
	 */
	protected void setSize(Endpoint stream, int width, int height)
	{
		Log.v(TAG, "Stream frame size has changed: w(" + width + "), h(" + height + ")");
		
		// Generate 3 texture ids for Y/U/V and place them into |textures|,
		// allocating enough storage for |width|x|height| pixels.
		int[] textures = yuvTextures[0];
		GLES20.glGenTextures(3, textures, 0);
		final int	halfWidth = width / 2;
		final int	halfHeight = height / 2;
		for ( int i=0; i<3; ++i )
		{
			int	w = (i == 0) ? width : halfWidth;
			int h = (i == 0) ? height : halfHeight;
			GLES20.glActiveTexture(GLES20.GL_TEXTURE0 + i);
			GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textures[i]);
			GLES20.glTexImage2D(GLES20.GL_TEXTURE_2D, 0, GLES20.GL_LUMINANCE, w, h, 0,
			                    GLES20.GL_LUMINANCE, GLES20.GL_UNSIGNED_BYTE, null);
			GLES20.glTexParameterf(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MIN_FILTER,
			                       GLES20.GL_LINEAR);
			GLES20.glTexParameterf(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MAG_FILTER,
			                       GLES20.GL_LINEAR);
			GLES20.glTexParameterf(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_S,
			                       GLES20.GL_CLAMP_TO_EDGE);
			GLES20.glTexParameterf(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_T,
			                       GLES20.GL_CLAMP_TO_EDGE);
		}
		checkNoGLES2Error();

		// Inform the listener, if there is one, that the frame dimensions have
		// changed
		if (mListener != null)
		{
			mListener.onFrameSizeChanged(width, height, stream, this);
		}
	}

	private Point					screenDimensions;
	private final int[][]			yuvTextures = {{-1, -1, -1}};
	private int						posLocation = -1;
	private final FramePool			framePool = new FramePool();
	private long					numFramesSinceLastLog = 0;
	private long					lastFPSLogTime = System.nanoTime();
	protected VideoSurfaceListener	mListener;
	private volatile boolean 		m_paused = false;
	
	/**
	 * Poor-man's assert(): die with |msg| unless |condition| is true.
	 */
	private static void abortUnless(boolean condition, String msg)
	{
		if ( !condition )
			throw new RuntimeException(msg);
	}

	/**
	 * Assert that no OpenGL ES 2.0 error has been raised.
	 */
	private static void checkNoGLES2Error()
	{
		int error = GLES20.glGetError();
		abortUnless(error == GLES20.GL_NO_ERROR, "GLES20 error: " + error);
	}

	/**
	 * Wrap a float[] in a direct FloatBuffer using native byte order.
	 * @param array
	 * @return
	 */
	private static FloatBuffer directNativeFloatBuffer(float[] array)
	{
		FloatBuffer	buffer = ByteBuffer.allocateDirect(array.length * 4)
								.order(ByteOrder.nativeOrder()).asFloatBuffer();
		buffer.put(array);
		buffer.flip();
		return buffer;
	}

	/**
	 * Compile & attach a |type| shader specified by |source| to |program|.
	 * @param type
	 * @param source
	 * @param program
	 */
	private static void addShaderTo(int type, String source, int program)
	{
		int[]	result = new int[]{GLES20.GL_FALSE};
		int		shader = GLES20.glCreateShader(type);
		GLES20.glShaderSource(shader, source);
		GLES20.glCompileShader(shader);
	    GLES20.glGetShaderiv(shader, GLES20.GL_COMPILE_STATUS, result, 0);
	    abortUnless(result[0] == GLES20.GL_TRUE,
	    		GLES20.glGetShaderInfoLog(shader) + ", source: " + source);
	    GLES20.glAttachShader(program, shader);
	    GLES20.glDeleteShader(shader);
	    checkNoGLES2Error();
    }

	/**
	 * Constructor.
	 *
	 * @param c
	 * @param screenDimensions
	 * @param listener
	 */
	public VideoSurfaceImpl(Context c, Point screenDimensions, VideoSurfaceListener listener)
	{
		super(c);
		Log.d(TAG, "Constructor");
		this.screenDimensions = screenDimensions;
		mListener = listener;
		setEGLContextClientVersion(2); // Pick an OpenGL ES 2.0 context.
		setRenderer(this);
		setRenderMode(RENDERMODE_WHEN_DIRTY);
		m_paused = false;
	}
	
	public void setListener(VideoSurfaceListener listener) {
		this.mListener = listener;
	}

	/**
	 * Queue |frame| to be uploaded.
	 * @param frame
	 */
	void queueFrame(I420Frame frame)
	{
		/* Paying for the copy of the YUV data here allows CSC and painting time to get spent on
		 * the render thread instead of the UI thread.
		 */
		abortUnless(FramePool.validateDimensions(frame), "Frame too large!");
		final I420Frame	newframe = framePool.takeFrame(frame);
		
		if (m_paused)
		{
			//Log.d("MOTION", "queueFrame no");
			
			return;
		}
		
		if (newframe != null)
		{
		    final I420Frame	frameCopy = newframe.copyFrom(frame);
		    queueEvent(new Runnable()
		    {
		    	public void run()
		    	{
                    if(setSizeParams != null && currentContextExists()) {
                        setSize(setSizeParams.endpoint, setSizeParams.width, setSizeParams.height);
                        setSizeParams = null;
                    }

		    		updateFrame(frameCopy);
		        }
	    	});
		}
    }

	/**
	 * Upload the planes from |frame| to the textures owned by this View.
	 * @param frame
	 */
	protected void updateFrame(I420Frame frame)
	{
		int[]	textures = yuvTextures[0];
	    texImage2D(frame, textures);
	    framePool.returnFrame(frame);
	 
	    if (!m_paused)
	    	requestRender();
    }

	@Override
	public void onDrawFrame(GL10 gl)
	{
	    GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT);
	    drawRectangle(yuvTextures[0], vertices);
	    ++numFramesSinceLastLog;
	    long	now = System.nanoTime();
	    if ( (lastFPSLogTime == -1) || ((now - lastFPSLogTime) > 1e9) )
	    {
	    	double	fps = numFramesSinceLastLog / ((now - lastFPSLogTime) / 1e9);
	    	Log.d(TAG, "Rendered FPS: " + fps);
	    	lastFPSLogTime = now;
	    	numFramesSinceLastLog = 1;
	    }
	    checkNoGLES2Error();
	}

	@Override
	protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec)
	{
		// this method responsible for determining the size of this view and its 
		// contents, and storing the details via setMesauredDimensions. 
		
		// If the size has not been specified explicitly via the constructor
		if(screenDimensions == null) {
			// Use the attributes from the XML
			int width  = View.MeasureSpec.getSize(widthMeasureSpec);
			int height = View.MeasureSpec.getSize(heightMeasureSpec);
			setMeasuredDimension(width, height);		    
		} else {
			// Else use the explicit values
			setMeasuredDimension(screenDimensions.x, screenDimensions.y);
		}
	}

	@Override
	public void onSurfaceChanged(GL10 gl, int width, int height)
	{
		Log.d(TAG, "onSurfaceChanged " + width + " - " + height);
		this.width = width;
		this.height = height;
		setDimensions(new Point(width, height));
	    GLES20.glViewport(0, 0, width, height);
	    checkNoGLES2Error();
	}

	@Override
	public void onSurfaceCreated(GL10 gl, EGLConfig config)
	{
		Log.d(TAG, "onSurfaceCreated");
	    int program = GLES20.glCreateProgram();
	    addShaderTo(GLES20.GL_VERTEX_SHADER, VERTEX_SHADER_STRING, program);
	    addShaderTo(GLES20.GL_FRAGMENT_SHADER, FRAGMENT_SHADER_STRING, program);

	    GLES20.glLinkProgram(program);
	    int[] result = new int[] { GLES20.GL_FALSE };
	    result[0] = GLES20.GL_FALSE;
	    GLES20.glGetProgramiv(program, GLES20.GL_LINK_STATUS, result, 0);
	    abortUnless(result[0] == GLES20.GL_TRUE,
	        GLES20.glGetProgramInfoLog(program));
	    GLES20.glUseProgram(program);

	    GLES20.glUniform1i(GLES20.glGetUniformLocation(program, "y_tex"), 0);
	    GLES20.glUniform1i(GLES20.glGetUniformLocation(program, "u_tex"), 1);
	    GLES20.glUniform1i(GLES20.glGetUniformLocation(program, "v_tex"), 2);

	    // Actually set in drawRectangle(), but queried only once here.
	    posLocation = GLES20.glGetAttribLocation(program, "in_pos");

	    int tcLocation = GLES20.glGetAttribLocation(program, "in_tc");
	    GLES20.glEnableVertexAttribArray(tcLocation);
	    GLES20.glVertexAttribPointer(
	        tcLocation, 2, GLES20.GL_FLOAT, false, 0, textureCoords);

	    GLES20.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	    checkNoGLES2Error();
	    
	    if (mListener != null)
	    {
	        mListener.onSurfaceRenderingStarted(this);
	    }
	}

	   @Override
	   public void onPause() 
	   {
		   super.onPause();
		   Log.d(TAG, "onPause");
		   m_paused = true;
	   }

        public void destroy() {
            Log.d(TAG, "destroy");
            if(framePool != null) {
                framePool.destroy();
            }
        }

	    @Override
	    public void onResume() 
	    {
	    	super.onResume();
			Log.d(TAG, "onPause");
			m_paused = false;
	    }
	
	/**
	 * Draw |textures| using |vertices| (X,Y coordinates).
	 * @param textures
	 * @param vertices
	 */
	private void drawRectangle(int[] textures, FloatBuffer vertices)
	{
		for ( int i=0; i<3; ++i )
		{
			GLES20.glActiveTexture(GLES20.GL_TEXTURE0 + i);
			GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textures[i]);
		}

		GLES20.glVertexAttribPointer(posLocation, 2, GLES20.GL_FLOAT, false, 0, vertices);
		GLES20.glEnableVertexAttribArray(posLocation);

	    GLES20.glDrawArrays(GLES20.GL_TRIANGLE_STRIP, 0, 4);
	    checkNoGLES2Error();
    }

	/**
	 * Upload the YUV planes from |frame| to |textures|.
	 * @param frame
	 * @param textures
	 */
	private void texImage2D(I420Frame frame, int[] textures)
	{
		final int	halfWidth = frame.width / 2;
		final int	halfHeight = frame.height / 2;
		for ( int i=0; i<3; ++i )
		{
			ByteBuffer	plane = frame.yuvPlanes[i];
			GLES20.glActiveTexture(GLES20.GL_TEXTURE0 + i);
			GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textures[i]);
			int	w = (i == 0) ? frame.width : halfWidth;
			int	h = (i == 0) ? frame.height : halfHeight;
			abortUnless(w == frame.yuvStrides[i], frame.yuvStrides[i] + "!=" + w);
			GLES20.glTexImage2D(GLES20.GL_TEXTURE_2D, 0, GLES20.GL_LUMINANCE, w, h, 0,
			                    GLES20.GL_LUMINANCE, GLES20.GL_UNSIGNED_BYTE, plane);
		}
		checkNoGLES2Error();
	}

	/**
	 *
	 * @param stream
	 * @return
	 */
	public VideoRendererCallbacks createCallbacks(Endpoint stream)
	{
		return new VideoRendererCallbacks(this, stream);
	}

	@Override
	public void setDimensions(Point dimensions)
	{
		screenDimensions = dimensions;
		
		// Run on UI thread
		new Handler(Looper.getMainLooper()).post(new Runnable() {
			@Override
			public void run() {
				requestLayout();		
			}
		});
	}
	
	private static class SetSizeParameters {
		final Endpoint endpoint;
		final int width;
		final int height;

		public SetSizeParameters(Endpoint endpoint, int width, int height) {
			this.endpoint = endpoint;
			this.width = width;
			this.height = height;
		}
	}
}
